# -*- coding: utf-8 -*-
"""
 Created on Tue Jul  7 14:24:09 2020

@author: joseph-hellerstein

A model study is an application of a model to multiple sets of
observed data (fitting the same model parameters). We refer to each
collection of observed data as a data source.
ModelStudy provides a way to create one ModelFitter for each
data source, and provides access to these fitters.


data_sets should be a list of data inputs as required by ModelFitter.

Usage
-----
    modelStudy = ModelStudy(model, dataSources, parameters)
    modelStudy.bootstrap()  # Create and save the bootstrap results
    modelStudy.plotFitAll()  # Plot of observed and fitted values for the study
    modelstudy.plotParameterEstimates()
"""

from SBstoat.namedTimeseries import NamedTimeseries, TIME
import SBstoat._plotOptions as po
from SBstoat.logs import Logger
from SBstoat.modelFitter import ModelFitter

from docstring_expander.expander import Expander
import inspect
import matplotlib.pylab as plt
import numpy as np
import os
import pandas as pd
import warnings


DIR_NAME = "ModelStudyFitters"
MIN_COUNT_BOOTSTRAP = 10


############## FUNCTIONS ###################
def mkDataSourceDct(filePath, colName, dataSourceNames=None, isTimeColumns=True):
    """
    Creates a dataSource dictionary as required by ModelStudy from
    a file whose columns are observed values of the same variable.

    Parameters
    ----------
    filepath: str
        path to the file containing columns of observed values
    colName: str
        Name of the simulation variable to fit to observed values
    dataSourceNames: list-str
        Names for the instances of observedValues
        if None, then column headers are used (or column number)
    isTimeColumns: boolean
        Columns are time. If False, rows are time.

    Returns
    -------
    dict: key is instance name; value is NamedTimeseries
    """
    dataDF = pd.read_csv(filePath, header=None)
    if isTimeColumns:
        dataDF = dataDF.transpose()
    if dataSourceNames is not None:
        # Use instance names
        if len(dataSourceNames) != len(dataDF.columns):
            msg = "Number of instances is not equal to the number of columns"
            raise ValueError(msg)
        dct = {k: v for k, v in zip(dataDF.columns, dataSourceNames)}
        dataDF = dataDF.rename(columns=dct)
    # Ensure that column names are strings
    dct = {k: str(k) for k in dataDF.columns}
    dataDF = dataDF.rename(columns=dct)
    dataDF.index.name = "time"
    # Construct the data source dictionary
    dataDF.index = range(len(dataDF))
    dataDF.index.name = TIME
    dataTS = NamedTimeseries(dataframe=dataDF)
    dataSourceDct = {d: dataTS.subsetColumns([d]) for d in dataDF.columns}
    dataSourceDct = {d: dataSourceDct[d].rename(d, colName)
          for d in dataSourceDct.keys()}
    #
    return dataSourceDct


############## CLASSES ###################
class ModelStudy():

    def __init__(self, modelSpecification, dataSources,
          dirStudyPath=None,
          instanceNames=None,
          logger=Logger(),
          useSerialized=True, doSerialize=True,
          isPlot=True,  **kwargs):
        """
        Parameters
        ---------
        modelSpecification: ExtendedRoadRunner/str
            roadrunner model or antimony model
        dataSources: list-NamedTimeseries/list-str or dict of either
            str: path to CSV file
        dirStudyPath: str
            Path to the output directory containing the serialized fitters
            for the study.
        instanceNames: list-str
            Names of study instances corresponds to the list of dataSources
        useSerialized: bool
            Use the serialization of each ModelFitter, if it exists
        doSerialized: bool
            Serialize each ModelFitter
        isPlot: bool
            Do plots
        kwargs: dict
            arguments passed to ModelFitter
        """
        self.dirStudyPath = dirStudyPath  # Path to the directory serialized ModelFitter
        if self.dirStudyPath is None:
            length = len(inspect.stack())
            absPath = os.path.abspath((inspect.stack()[length-1])[1])
            dirCaller = os.path.dirname(absPath)
            self.dirStudyPath = os.path.join(dirCaller, DIR_NAME)
        self.isPlot = isPlot
        self.doSerialize = doSerialize
        self.useSerialized = useSerialized
        self.instanceNames, self.dataSourceDct = self._mkInstanceData(
              instanceNames, dataSources)
        self.fitterPathDct = {}  # Path to serialized fitters; key is instanceName
        self.fitterDct = {}  # Fitters: key is instanceName
        self.logger = logger
        # Ensure that the directory exists
        if not os.path.isdir(self.dirStudyPath):
            os.makedirs(self.dirStudyPath)
        # Construct the fitters
        for name, dataSource in self.dataSourceDct.items():
            filePath = self._getSerializePath(name)
            self.fitterPathDct[name] = filePath
            if os.path.isfile(filePath) and useSerialized:
                self.fitterDct[name] = ModelFitter.deserialize(filePath)
            else:
                self.fitterDct[name] = ModelFitter(modelSpecification,
                       dataSource, logger=self.logger,
                       isPlot=self.isPlot, **kwargs)
                self._serializeFitter(name)

    def _mkInstanceData(self, instanceNames, dataSources):
        # Determine value for instances of observed data
        if isinstance(instanceNames, list):
            newInstanceNames = list(instanceNames)
        elif instanceNames is None:
            newInstanceNames = ["src_%d" %d
                  for d in range(1, len(dataSources)+1)]
        else:
            msg = "Invalid type for instanceNames: %s" % str(type(instanceNames))
            msg += "  \nMust be list or None."
            raise ValueError(msg)
        # Value of newDataSourceDct
        if isinstance(dataSources, dict):
            newDataSourceDct = dict(dataSources)
            newInstanceNames = list(newDataSourceDct.keys())
        elif isinstance(dataSources, list):
            newDataSourceDct = {n: dataSources[i]
                  for i, n in enumerate(newInstanceNames)}
        else:
            msg = "Invalid type for dataSources: %s" % str(type(dataSources))
            msg += "  \nMust be dict or list."
            raise ValueError(msg)
        #
        return newInstanceNames, newDataSourceDct

    def _getSerializePath(self, name):
        """
        Constructs to serialization file for fitter.

        Parameters
        ----------
        name: str
            Name of the fitter instance/data source

        Returns
        -------
        str
        """
        return os.path.join(self.dirStudyPath, "%s.pcl" % name)

    def _serializeFitter(self, name):
        filePath = self._getSerializePath(name)
        if self.doSerialize:
            self.fitterDct[name].serialize(filePath)

    def fitModel(self):
        """
        Does fits for all models and serializes the results.
        """
        for name in self.instanceNames:
            self.logger.activity("Fit for data %s" % name)
            fitter = self.fitterDct[name]
            fitter.fitModel()
            self._serializeFitter(name)
            self.logger.result(fitter.reportFit())

    def _hasBootstrapResult(self, fitter):
        result = False
        if fitter.bootstrapResult is not None:
            if fitter.bootstrapResult.numSimulation > 0:
                result = True
        return result

    def bootstrap(self, **kwargs):
        """
        Does bootstrap for models that have not bootstrap. Serializes the result.

        Parameters
        ----------
        kwargs: dict
            arguments passed to ModelFitter.bootstrap
        """
        for name in self.instanceNames:
            fitter = self.fitterDct[name]
            if (not self.useSerialized) and (not self._hasBootstrapResult(fitter)):
                msg = "Doing bootstrapp for instance %s" % name
                self.logger.activity(msg)
                if fitter.params is None:
                    fitter.fitModel()
                fitter.bootstrap(**kwargs)
                if not self._hasBootstrapResult(fitter):
                    # Not a valid bootstrap result
                    fitter.bootstrapResult = None
                self._serializeFitter(name)
            else:
                if self._hasBootstrapResult(fitter):
                    msg = "Using existing bootstrap results (%d) for %s"  \
                          % (fitter.bootstrapResult.numSimulation, name)
                    self.logger.result(msg)
                else:
                    msg = "No bootstrap results for data source %s"  % name
                    self.logger.result(msg)

    @Expander(po.KWARGS, po.BASE_OPTIONS, indent=8, header=po.HEADER)
    def plotFitAll(self, **kwargs):
        """
        Constructs plots using parameters from bootstrap if available.

        Parameters
        ----------
        #@expand
        """
        for name in self.instanceNames:
            newKwargs = dict(kwargs)
            if po.SUPTITLE in newKwargs.keys():
                title = "%s: %s" % (name, newKwargs[po.SUPTITLE])
            else:
                title = "%s: Fitted vs. Observed (with 95th percentile)"  \
                      % name
            newKwargs[po.SUPTITLE] = title
            fitter = self.fitterDct[name]
            if fitter.params is None:
                self.logger.result("Must do fitModel or bootstrap before plotting.")
            else:
                if fitter.bootstrapResult is not None:
                    if fitter.bootstrapResult.numSimulation > 0:
                        fitter.plotFitAll(**newKwargs)

    @Expander(po.KWARGS, po.BASE_OPTIONS, indent=8, header=po.HEADER)
    def plotParameterEstimates(self):
        """
        Parameters
        ----------
        #@expand
        """
        SCALE = 1.1  # Amount by which to scale an upper boundary
        fitterDct = {}
        for dataSource, fitter in self.fitterDct.items():
            if self._hasBootstrapResult(fitter):
                for parameterName, values in  \
                      fitter.bootstrapResult.parameterDct.items():
                    length = len(values)
                    if length < MIN_COUNT_BOOTSTRAP:
                        msg = "Only %d samples from bootstrap of %s."  \
                              % (length, dataSource)
                        msg += "Unable to do plot for parameter %s."  \
                               % parameterName
                        self.logger.result(msg)
                    else:
                        fitterDct[dataSource] = fitter
        if len(fitterDct) == 0:
            self.logger.result("No data to plot.")
        else:
            instanceNames = fitterDct.keys()
            trues = [f.bootstrapResult is None for f in fitterDct.values()]
            if any(trues):
                raise ValueError("\n***Must do bootstrap before getting report.")
            fitter = list(fitterDct.values())[0]
            parameters = fitter.parametersToFit
            parameterDct = fitter.getDefaultParameterValues()
            # Construct plot
            _, axes = plt.subplots(len(parameters),1, figsize=(12,10))
            for pos, pName in enumerate(parameters):
                ax = axes[pos]
                means = [f.bootstrapResult.parameterMeanDct[pName]
                      for f in fitterDct.values()]
                stds = [f.bootstrapResult.parameterStdDct[pName]
                      for f in fitterDct.values()]
                y_upper = max([s + m for s, m in zip(stds, means)])*SCALE
                y_upper = max(y_upper, parameterDct[pName]*SCALE)
                ax.errorbar(instanceNames, means, yerr=stds, linestyle="")
                ax.scatter(instanceNames, means, s=18.0)
                values = np.repeat(parameterDct[pName], len(instanceNames))
                ax.plot(instanceNames, values, linestyle="dashed", color="grey")
                ax.set_title(pName)
                ax.set_ylim([0, y_upper])
                if pos == len(parameters) - 1:
                # Ignore bogus warning from matplotlib
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        ax.set_xticklabels(instanceNames)
                    ax.set_xlabel("Observed Data")
                else:
                    ax.set_xticklabels([])
                    ax.set_xlabel("")
            suptitle = "Bootstrap Parameters With 1-Standard"
            suptitle += "\n(Dashed line is parameter value in model.)"
            _ = plt.suptitle(suptitle)
            if self.isPlot:
                plt.show()



"""
Calculation of statistics for time series. Usage:
    statistic = TimeseriesStatistic(baseTimeseries)
    statistic.accumulate(timeseries_1)
    statistic.accumulate(timeseries_2)
    statistic.calculate()
    print(statistic.count)  # Number of timeseries accumulated
    print(statistic.meanTS)  # Timeseries of mean values
    print(statistic.stdTS)  # Timeseries of std values
    print(statistic.lowerPercentileTS)  # Timeseries of lower 5% at each time
    print(statistic.upperPercentileTS)  # Timeseries of lower 95% at each time
"""

from SBstoat.namedTimeseries import NamedTimeseries
from SBstoat import rpickle

import copy
import numpy as np


PERCENTILES = [5.0, 50.0, 95.0]  # Percentiles calculated


class TimeseriesStatistic(rpickle.RPickler):

    def __init__(self, prototypeTS:NamedTimeseries, percentiles:list=None):
        """
        Parameters
        ----------
        prototypeTS: same length and columns as desired
        percentiles: percentiles to calculate for accumulated Timeseries
        """
        # Statistics
        if prototypeTS is not None:
            self.count = 0  # Count of timeseries accumulated
            self.prototypeTS = prototypeTS
            self.colnames = self.prototypeTS.colnames
            self.sumTS = self.prototypeTS.copy(isInitialize=True)
            self.ssqTS = self.prototypeTS.copy(isInitialize=True)
            if percentiles is None:
                percentiles = PERCENTILES
            self.percentiles = percentiles
            self._timeseries_list = []  # List of timeseries accumulated
            # Means
            self.meanTS = prototypeTS.copy(isInitialize=True) # means
            # Standard deviations
            self.stdTS = prototypeTS.copy(isInitialize=True)  # standard deviations
            # Percentiles
            self.percentileDct = {}  # Key: percentile; Value: Timeseries
        else:
            # rpConstruct initializations.
            pass

    @classmethod
    def rpConstruct(cls):
        """
        Overrides rpickler.rpConstruct to create a method that
        constructs an instance without arguments.

        Returns
        -------
        Instance of cls
        """
        return cls(None)

    def copy(self):
        """
        Makes a copy of the object, including internal state.

        Returns
        -------
        TimeseriesStatistic
        """
        newStatistic = TimeseriesStatistic(self.prototypeTS,
              percentiles=self.percentiles)
        # Update internal state
        for attr in self.__dict__.keys():
            if "__" in attr:
                continue
            obj = self.__getattribute__(attr)
            if "copy" in dir(obj):
                # Use native copy
                newObj = obj.copy()
            else:
                newObj = copy.deepcopy(obj)
            newStatistic.__setattr__(attr, newObj)
        return newStatistic

    def equals(self, other):
        """
        Checks if the two instances have the same values.

        Parameters
        ----------
        other: TimeseriesStatistic

        Returns
        -------
        bool
        """
        isEqual = True
        for attr in self.__dict__.keys():
            if "__" in attr:
                continue
            expression = "dir(self.%s)" % attr
            if "equals" in eval(expression):
                expression = "self.%s.equals(other.%s)" % (attr, attr)
            else:
                expression = "self.%s == other.%s" % (attr, attr)
            isEqual = isEqual and eval(expression)
        return isEqual

    def accumulate(self, newTS:NamedTimeseries):
        """
        Accumulates statistics for a new timeseries.
        """
        self.sumTS[self.colnames] = self.sumTS[self.colnames]  \
               + newTS[self.colnames]
        self.ssqTS[self.colnames] = self.ssqTS[self.colnames]  \
               + newTS[self.colnames]**2
        if len(self.percentiles) > 0:
            self._timeseries_list.append(newTS)
        self.count += 1

    def calculate(self):
        """
        Calculates statistics.
        """
        if self.count > 0:
            self._calculateMean()
        if self.count > 1:
            self._calculateStd()
            self._calculatePercentiles()

    def _calculateMean(self):
        for col in self.colnames:
            self.meanTS[col] = (1.0*self.sumTS[col]) / self.count

    def _calculateStd(self):
        for col in self.colnames:
            if self.count > 1:
                self.stdTS[col] = self.ssqTS[col] - self.count*self.meanTS[col]**2
                self.stdTS[col] = self.stdTS[col] / (self.count - 1)
                self.stdTS[col] = np.array([
                      0.0 if np.isclose(v, 0.0) else v for v in self.stdTS[col]])
                self.stdTS[col] = np.sqrt(self.stdTS[col])
            else:
                pass

    def _calculatePercentiles(self):
        if len(self.percentiles) == 0:
            return
        self.percentileDct = {}
        refTS = self._timeseries_list[0]
        indices = range(len(refTS))
        for percentile in self.percentiles:
            percentileTS = refTS.copy(isInitialize=True)
            for col in self.colnames:
                col_values = []
                for idx in indices:
                    time_values = [t[col][idx]  for t in self._timeseries_list]
                    col_values.append(np.percentile(time_values, percentile))
                percentileTS[col] = np.array(col_values)
            self.percentileDct[percentile] = percentileTS

    def _merge(self, others):
        """
        Merges a colection of TimeseriesStatistic for the same
        shape of timeseries.

        Parameters
        ----------
        others: list-TimeseriesStatistic

        Returns
        -------
        TimeseriesStatistic
        """
        result = self.copy()
        for other in others:
            result.count += other.count
            result.sumTS[result.colnames] += other.sumTS[result.colnames]
            result.ssqTS[result.colnames] += other.ssqTS[result.colnames]
            result._timeseries_list.extend(other._timeseries_list)
        return result

    @classmethod
    def merge(cls, others):
        """
        Merges a colection of TimeseriesStatistic for the same
        shape of timeseries.

        Parameters
        ----------
        others: list-TimeseriesStatistic

        Returns
        -------
        TimeseriesStatistic
        """
        statistic = others[0]
        newOthers = list(others[1:])
        return statistic._merge(newOthers)



"""Provides for incremental construction of python statements."""


class StatementManager():

    def __init__(self, func):
        """
        Parameters
        ----------
        func: Function
        """
        self.func = func
        self.pargs = []
        self.kwargs = {}

    def addKwargs(self, **kwargs):
        self.kwargs.update(kwargs)

    def addPosarg(self, arg):
        self.pargs.append(arg)

    def addPosargs(self, args):
        self.pargs.extend(args)

    def execute(self):
        return self.func(*self.pargs, **self.kwargs)



"""
Adapations of pickle to make it more robust to changes in pickled objects.

Provides robustness by not serializing the methods of the target object, only
its __dict__. Deserialization involves: (a) create a default instantiation
of the object and (b) re-assigning its __dict__.

Objects for which this behavior is desired must inherit RPickler and
override the class method RPicklerConstructor(cls) as appropriate to
so that the method returns a default instance of the class.

"""

import copy
import os
import pickle


class Serialization():
    """RPickler serialization of an object and its sub-objects."""

    def __init__(self, obj):
        """
        Parameters
        ----------
        obj: Object being serialized
        """
        self.cls = obj.__class__  # Class being serialized
        self.dct = dict(obj.__dict__)  # __dict__ for the instance

    def serialize(self):
        """
        Recursively constructs the serialization of the object.
        """
        for key, value in self.dct.items():
            if issubclass(value.__class__, RPickler):
                self.dct[key] = Serialization(value)
                self.dct[key].serialize()

    def deserialize(self):
        """
        Recursively deserializes objects.

        Returns
        -------
        self.cls
        """
        obj = self.cls.rpConstruct()
        # Recursively instantiate serialized objects.
        # Save as instances of the constructed object.
        for key, value in self.dct.items():
            if isinstance(value, Serialization):
                obj.__dict__[key] = value.deserialize()
            else:
                obj.__dict__[key] = copy.deepcopy(value)
        # Revise the obj as required
        obj.rpRevise()
        #
        return obj

    def __repr__(self):
        return "Serialization of %s" % str(self.cls)


class RPickler():
    # Used by classes that implement robust pickling.

    @classmethod
    def rpConstruct(cls):
        """
        Provides a default construction of an object.

        Returns
        -------
        Instance of cls
        """
        return cls()

    def rpRevise(self):
        """
        Provides a hook to modify instance variables after they have
        been initialized by RPickle.
        """


def dump(obj, fd):
    """
    Dumps the objects to a file.

    Parameters
    ----------
    obj: object to be serialized
    fd: file descriptor
    """
    # Construct the serialization
    serialization = Serialization(obj)
    serialization.serialize()
    # Serialize
    pickle.dump(serialization, fd)

def load(fd):
    """
    Restores a serialized object.

    Parameters
    ----------
    fd: file descriptor

    Returns
    -------
    object
    """
    serialization = pickle.load(fd)
    return serialization.deserialize()



# -*- coding: utf-8 -*-
"""
 Created on August 30, 2020

@author: joseph-hellerstein

Creates synthetic observations.

TODO:
  1. Remove outliers via std criteria
  2. Handling missing observerations in observedTS
"""

from SBstoat.namedTimeseries import NamedTimeseries
from SBstoat import _helpers

import abc
import numpy as np
import pandas as pd


class ObservationSynthesizer(abc.ABC):


    def __init__(self, observedTS:NamedTimeseries=None,
                 fittedTS:NamedTimeseries=None,
                 residualsTS:NamedTimeseries=None):
        """
        Child classes must specify two of the three timeseries.

        Parameters
        ----------
        observedTS: Observations
        fittedTS: Fitted values
        residualsTS: Residual values
        """
        def getTS(data):
            if isinstance(data, pd.DataFrame):
                return NamedTimeseries(dataframe=data)
            return data
        #
        self._observedTS = getTS(observedTS)
        self._fittedTS = getTS(fittedTS)
        self._residualsTS = getTS(residualsTS)
        if self._observedTS is not None:
            self._observedTS = self._observedTS.copy()
        if self._fittedTS is not None:
            self._fittedTS = self._fittedTS.copy()
        if self._residualsTS is not None:
            self._residualsTS = self._residualsTS.copy()
        # Set the columns based on the information provided
        for ts in [self._observedTS, self._fittedTS, self._residualsTS]:
            if ts is not None:
                self.columns = ts.colnames
                break
        np.random.seed()  # Ensure different sequences on each invocation

    @property
    def observedTS(self):
        if self._observedTS is None:
            self._observedTS = self._fittedTS.copy()
            self._observedTS[self.columns] += self._residualsTS[self.columns]
        return self._observedTS

    @property
    def fittedTS(self):
        if self._fittedTS is None:
            self._fittedTS = self._observedTS.copy()
            self._fittedTS[self.columns] -= self._residualsTS
        return self._fittedTS

    @property
    def residualsTS(self):
        if self._residualsTS is None:
            self._residualsTS = self._observedTS.copy()
            self._residualsTS[self.columns] -=  \
                self._fittedTS[self.columns]
            self._residualsTS[self.columns]  \
                  = np.nan_to_num(self._residualsTS[self.columns], nan=0.0)
        return self._residualsTS

    @abc.abstractmethod
    def calculate(self):
        """
        Calculates a set of synthetic observations
        """
        pass


class ObservationSynthesizerRandomizedResiduals(ObservationSynthesizer):

    def __init__(self, observedTS:NamedTimeseries=None,
                 fittedTS:NamedTimeseries=None,
                 residualsTS:NamedTimeseries=None,
                 filterSL:float=0.01):
        """
        Parameters
        ----------
        filterSL: maximum significance level used in filtering residuals

        Notes
        -----
        observedTS and fittedTS must be non-Null.
        """
        super().__init__(observedTS=observedTS, fittedTS=fittedTS,
                 residualsTS=residualsTS)
        self._filterSL = filterSL
        self._filteredResidualsDct = self._calcFilteredResidualsDct()

    def _calcFilteredResidualsDct(self):
        # Constructs dictionary of residuals meeting std criteria
        dct = {}
        for col in self._observedTS.colnames:
            if self._filterSL is None:
                dct[col] = self.residualsTS[col]
            else:
                dct[col] = _helpers.filterOutliersFromZero(
                      self.residualsTS[col], self._filterSL)
        return dct

    def calculate(self):
        """
        Calculates synthetic observations by randomly rearranging residuals.
        """
        numRow = len(self.observedTS)
        newObservedTS = self.fittedTS.copy()
        for column in self.columns:
            if len(self._filteredResidualsDct[column]) == 0:
                msg = "No residuals left after filtering."
                msg += " Make filter less restrictive."
                raise ValueError(msg)
            newObservedTS[column] += np.random.choice(
                self._filteredResidualsDct[column], numRow, replace=True)
        return newObservedTS


class ObservationSynthesizerRandomErrors(ObservationSynthesizer):
    """
    Calculates residuals using randomized errors from a normal
    distribution.
    """

    def __init__(self, observedTS=None,
              fittedTS:NamedTimeseries=None, std:float=0.1):
        """
        Parameters
        ----------
        std: standard deviation for random numbers
        """
        if fittedTS is None:
            raise ValueError("Must supply fitted timeseries")
        self.std = std
        super().__init__(fittedTS=fittedTS)

    def calculate(self):
        """
        Calculates synthetic observations by randomly rearranging residuals.
        """
        numRow = len(self.fittedTS)
        self._residualsTS = self._fittedTS.copy()
        for col in self.columns:
            self._residualsTS[col] = np.random.normal(
                0, self.std, numRow)
        return self.observedTS



"""
Harness used to test codes using SBML models.

Usage
-----
harness = TestHarness(sbmlPath, selectedColumns, parametersToFit)
harness.evaluate(stdResiduals=0.5, maxRelativeError=0.1)
"""

from SBstoat.modelFitter import ModelFitter
from SBstoat.namedTimeseries import NamedTimeseries, TIME
from SBstoat import _helpers
from SBstoat import logs
from SBstoat.observationSynthesizer import ObservationSynthesizerRandomErrors

import numpy as np
import tellurium as te
import typing


HTTP200 = 200
HTTP = "http://"
MAX_PARAMETER = 5  # Maximum number of parameters estimated
NUM_BOOTSTRAP_ITERATION = 100


class TestHarnessResult():

    def __init__(self):
        self.parameterRelErrorDct = {}

    def addParameter(self, name, relerror):
        self.parameterRelErrorDct[name] = relerror

    def __repr__(self):
        return str(self.parameterRelErrorDct)


class TestHarness():

    def __init__(self, sbmlPath:str,
          parametersToFit:typing.List[str]=None,
          selectedColumns:typing.List[str]=None,
          **kwargs):
        """
        Parameters
        ----------
        sbmlPath: path or URL to SBML files
        selectedColumns: names of floating species used in fitting
        parametersToFit: names of parameters to fit
        kwargs: passed to ModelFitter constructor
        """
        if "logger" in kwargs.keys():
            self.logger = kwargs["logger"]
        else:
            self.logger = logs.Logger()
        #
        self.sbmlPath = sbmlPath
        self.roadRunner = self._initializeRoadrunner()
        self.selectedColumns = selectedColumns
        self.parametersToFit = self._getSetableParameters(parametersToFit)
        self.kwargs = kwargs
        self.parametersToFit = self._getSetableParameters(parametersToFit)
        self.parameterValueDct = {p: self.roadRunner[p]
              for p in self.parametersToFit}
        self.fitModelResult = TestHarnessResult()
        self.bootstrapResult = TestHarnessResult()
        self._validate()

    def _getSetableParameters(self, initialParametersToFit,
          maxParameter=MAX_PARAMETER):
        if initialParametersToFit is None:
            parameters = list(self.roadRunner.getGlobalParameterIds())
        else:
            parameters = initialParametersToFit
        # See which parameters can be changed
        parametersToFit = []
        for parameter in parameters:
            try:
                self.roadRunner.model[parameter] = self.roadRunner.model[parameter]
                parametersToFit.append(parameter)
            except Exception:
                pass
        #
        if len(parametersToFit) > maxParameter:
            parametersToFit = parametersToFit[:maxParameter]
        return parametersToFit

    def _checkNamesInModel(self, names:typing.List[str], errorMsgPattern:str):
        """
        Verifies that the list of names is in the model.

        Parameters
        ----------
        names: what to check
        errorMsgPattern: string pattern taking one argument (name) for error msg

        Raises ValueError
        """
        diffs = set(names).difference(self.roadRunner.model.keys())
        if len(diffs) > 0:
            raise ValueError(errorMsgPattern % str(diffs))

    def _initializeRoadrunner(self):
        """
        Initializes the object and does error checking.

        Return
        ------
        Roadrunner
        """
        try:
            _ = te.loads(self.sbmlPath)
        except Exception as err:
            msg = "sbmlPath is not a valid file path or URL: %s" \
                  % self.sbmlPath
            self.logger.error("_initializeRoadrunner", err)
            raise err
        return te.loads(self.sbmlPath)

    def _validate(self):
        """
        Validates names of species and parameters
        """
        # Validate the column names
        errorMsgPattern = "Variable name is not in model: %s"
        if self.selectedColumns is not None:
            self._checkNamesInModel(self.selectedColumns, errorMsgPattern)
        # Validate the parameter names
        errorMsgPattern = "Parameter name is not in model: %s"
        self._checkNamesInModel(self.parametersToFit, errorMsgPattern)

    def _recordResult(self, params, relError, testHarnessResult):
        valuesDct = params.valuesdict()
        for name in self.parameterValueDct.keys():
            estimatedValue = valuesDct[name]
            actualRelError = _helpers.calcRelError(self.parameterValueDct[name],
                  estimatedValue)
            testHarnessResult.addParameter(name, actualRelError)
            if actualRelError > relError:
                self.logger.result("Parameter %s has high relError: %2.3f"
                      % (name, actualRelError))

    def evaluate(self, stdResiduals:float=0.1, relError:float=0.1,
          endTime:float=10.0, numPoint:int=30,
          fractionParameterDeviation:float=0.5,
          numIteration=NUM_BOOTSTRAP_ITERATION):
        """
        Evaluates model fitting accuracy and bootstrapping for model

        Parameters
        ----------
        stdResiduals: Standard deviations of variable used in constructing reiduals
        relError: relative error of parameter used in evaluation
        endTime: ending time of the simulation
        numPoint: number of points in the simulatin
        fractionParameterDeviation: fractional amount that the parameter can vary
        """
        # Construct synthetic observations
        if self.selectedColumns is None:
            data = self.roadRunner.simulate(0, endTime, numPoint)
        else:
            allColumns = list(self.selectedColumns)
            if TIME not in allColumns:
                allColumns.append(TIME)
            data = self.roadRunner.simulate(0, endTime, numPoint, allColumns)
        bracketTime = "[%s]" % TIME
        if bracketTime in data.colnames:
            # Exclude any column named '[time]'
            idx = data.colnames.index(bracketTime)
            dataArr = np.delete(data, idx, axis=1)
            colnames = list(data.colnames)
            colnames.remove(bracketTime)
            colnames = [s[1:-1] if s != TIME else s for s in colnames]
            simTS = NamedTimeseries(array=dataArr, colnames=colnames)
        else:
            simTS = NamedTimeseries(namedArray=data)
        synthesizer = ObservationSynthesizerRandomErrors(
              fittedTS=simTS, std=stdResiduals)
        observedTS = synthesizer.calculate()
        # Construct the parameter ranges
        parameterDct = {}
        for name in self.parameterValueDct.keys():
            lower = self.parameterValueDct[name]*(1 - fractionParameterDeviation)
            upper = self.parameterValueDct[name]*(1 + fractionParameterDeviation)
            value = np.random.uniform(lower, upper)
            parameterDct[name] = (lower, upper, value)
        # Create the fitter
        fitter = ModelFitter(self.roadRunner, observedTS,
              selectedColumns=self.selectedColumns, parameterDct=parameterDct,
              **self.kwargs)
        msg = "Fitting the parameters %s" % str(self.parameterValueDct.keys())
        self.logger.result(msg)
        # Evaluate the fit
        fitter.fitModel()
        self._recordResult(fitter.params, relError, self.fitModelResult)
        # Evaluate bootstrap
        fitter.bootstrap(numIteration=numIteration)
        if fitter.bootstrapResult is not None:
            if fitter.bootstrapResult.numSimulation > 0:
                self._recordResult(fitter.bootstrapResult.params,
                      relError, self.bootstrapResult)



"""
 Created on Aug 20, 2020

@author: joseph-hellerstein

Codes that provide various analyses of residuals.

There are 3 types of timeseries: observed, fitted, and residuals
(observed - fitted).

Plots are organized by the timeseries and the characteristic analyzed. These
characteristics are: (a) over time, (b) histogram.
"""

from SBstoat.namedTimeseries import NamedTimeseries
import SBstoat._plotOptions as po
from SBstoat import timeseriesPlotter as tp

from docstring_expander.expander import Expander
import numpy as np

PLOT = "plot"


class ResidualsAnalyzer():

    def __init__(self, observedTS:NamedTimeseries, fittedTS:NamedTimeseries,
              residualsTS:NamedTimeseries=None, meanFittedTS=None,
              stdFittedTS=None,
              bandLowTS:NamedTimeseries=None, bandHighTS:NamedTimeseries=None,
              isPlot:bool=True):
        """
        Parameters
        ----------
        observedTS: Observed values
        residualsTS: same time values as observedTS
        fittedTS: fitted values
            may have different times than observedTS
        meanFittedTS: fitted values with same times as observedTS
        stdFittedTS: fitted values with same times as observedTS
        bandLowTS: timeseries that describes the lower band for timeseries1
        bandhighTS: timeseries that describes the higher band for timeseries1
        """
        self.observedTS = observedTS
        self.fittedTS = fittedTS
        self.meanFittedTS = meanFittedTS
        self.stdFittedTS = stdFittedTS
        self.bandLowTS = bandLowTS
        self.bandHighTS = bandHighTS
        if residualsTS is None:
            self.residualsTS = self.observedTS.copy()
            cols = self.residualsTS.colnames
            self.residualsTS[cols] -= self.fittedTS[cols]
            self.residualsTS[cols]  \
                  = np.nan_to_num(self.residualsTS[cols], nan=0.0)
        else:
            self.residualsTS = residualsTS.copy()
        ### Plotter
        self._plotter = tp.TimeseriesPlotter(isPlot=isPlot)

    @staticmethod
    def _addKeyword(kwargs:dict, key:str, value:object):
        if not key in kwargs:
            kwargs[key] = value

    @Expander(po.KWARGS, po.BASE_OPTIONS, indent=8,
          header=po.HEADER)
    def plotAll(self, **kwargs:dict):
        """
        Does all residual plots.

        Parameters
        ----------
        #@expand
        """
        for name in dir(self):
            if name == "plotAll":
                continue
            if name[0:4] == PLOT:
                statement = "self.%s(**kwargs)" % name
                exec(statement)

    @Expander(po.KWARGS, po.BASE_OPTIONS, indent=8,
          header=po.HEADER)
    def plotResidualsOverTime(self, **kwargs:dict):
        """
        Plots residuals of a fit over time.

        Parameters
        ----------
        #@expand
        """
        ResidualsAnalyzer._addKeyword(kwargs, po.MARKER, "o")
        ResidualsAnalyzer._addKeyword(kwargs, po.SUPTITLE, "Residuals Over Time")
        self._plotter.plotTimeSingle(self.residualsTS, **kwargs)

    @Expander(po.KWARGS, po.BASE_OPTIONS, indent=8,
          header=po.HEADER)
    def plotFittedObservedOverTime(self, **kwargs:dict):
        """
        Plots the fit with observed data over time.

        Parameters
        ----------
        #@expand
        """
        title = "Observed vs. fitted"
        if self.bandLowTS is not None:
            title += " (with shading for 95th percentile)"
        ResidualsAnalyzer._addKeyword(kwargs, po.SUPTITLE, title)
        ResidualsAnalyzer._addKeyword(kwargs, po.MARKER, [None, "o", "^"])
        legends = ["fitted", "observed"]
        if self.meanFittedTS is not None:
            legends.append("bootstrap fitted")
        ResidualsAnalyzer._addKeyword(kwargs, po.LEGEND, legends)
        ResidualsAnalyzer._addKeyword(kwargs, po.COLOR, ["b", "b", "r"])
        self._plotter.plotTimeSingle(
              self.fittedTS,
              timeseries2=self.observedTS,
              meanTS=self.meanFittedTS, stdTS=self.stdFittedTS,
              bandLowTS=self.bandLowTS,
              bandHighTS=self.bandHighTS,
              **kwargs)

    @Expander(po.KWARGS, po.BASE_OPTIONS, includes=[po.BINS], indent=8,
          header=po.HEADER)
    def plotResidualsHistograms(self, **kwargs:dict):
        """
        Plots histographs of parameter values from a bootstrap.

        Parameters
        ----------
        parameters: List of parameters to do pairwise plots
        #@expand
        """
        ResidualsAnalyzer._addKeyword(kwargs, po.SUPTITLE, "Residual Distributions")
        self._plotter.plotHistograms(self.residualsTS, **kwargs)

    @Expander(po.KWARGS, po.BASE_OPTIONS, indent=8,
          header=po.HEADER)
    def plotResidualsAutoCorrelations(self, **kwargs:dict):
        """
        Plots auto correlations between residuals of columns.

        Parameters
        ----------
        parameters: List of parameters to do pairwise plots
        #@expand
        """
        ResidualsAnalyzer._addKeyword(kwargs, po.SUPTITLE, "Residual Autocorrelations")
        self._plotter.plotAutoCorrelations(self.residualsTS, **kwargs)

    @Expander(po.KWARGS, po.BASE_OPTIONS, indent=8,
          header=po.HEADER)
    def plotResidualCrossCorrelations(self, **kwargs:dict):
        """
        Plots cross correlations between residuals of columns.

        Parameters
        ----------
        parameters: List of parameters to do pairwise plots
        #@expand
        """
        ResidualsAnalyzer._addKeyword(kwargs, po.SUPTITLE, "Residual Cross Correlations")
        self._plotter.plotCrossCorrelations(self.residualsTS, **kwargs)



# -*- coding: utf-8 -*-
"""
 Created on August 18, 2020

@author: joseph-hellerstein

Core logic of model fitter. Does not include plots.
"""

from SBstoat.namedTimeseries import NamedTimeseries, TIME, mkNamedTimeseries
from SBstoat.logs import Logger
import SBstoat.timeseriesPlotter as tp
from SBstoat import rpickle

import collections
import copy
import lmfit
import numpy as np
import tellurium as te
import typing

# Constants
PARAMETER_LOWER_BOUND = 0
PARAMETER_UPPER_BOUND = 10
#  Minimizer methods
METHOD_DIFFERENTIAL_EVOLUTION = "differential_evolution"
METHOD_BOTH = "both"
METHOD_LEASTSQ = "leastsq"
METHOD_FITTER_DEFAULTS = [METHOD_DIFFERENTIAL_EVOLUTION, METHOD_LEASTSQ]
METHOD_BOOTSTRAP_DEFAULTS = [METHOD_LEASTSQ,
      METHOD_DIFFERENTIAL_EVOLUTION, METHOD_LEASTSQ]
MAX_CHISQ_MULT = 5
PERCENTILES = [2.5, 97.55]  # Percentile for confidence limits
INDENTATION = "  "
NULL_STR = ""
IS_REPORT = False
LOWER_PARAMETER_MULT = 0.95
UPPER_PARAMETER_MULT = 0.95
LARGE_RESIDUAL = 1000000


_BestParameters = collections.namedtuple("_BestParameters",
      "params rssq")  #  parameters, residuals sum of squares



##############################
class ParameterSpecification():

    def __init__(self, lower=None, value=None, upper=None):
        self.lower = lower
        self.value = value
        self.upper = upper


class ModelFitterCore(rpickle.RPickler):

    # Subclasses used in interface
    class OptimizerMethod():

        def __init__(self, method, kwargs):
            self.method = method
            self.kwargs = kwargs


    def __init__(self, modelSpecification, observedData,
          parametersToFit=None,
          selectedColumns=None,
          fitterMethods=None,
          numFitRepeat=1,
          bootstrapMethods=None,
          parameterLowerBound=PARAMETER_LOWER_BOUND,
          parameterUpperBound=PARAMETER_UPPER_BOUND,
          parameterDct=None,
          fittedDataTransformDct=None,
          logger=Logger(),
          isPlot=True,
          _loggerPrefix="",
          # The following must be kept in sync with ModelFitterBootstrap.bootstrap
          numIteration:int=10,
          reportInterval:int=1000,
          maxProcess:int=None,
          serializePath:str=None,
          ):
        """
        Constructs estimates of parameter values.

        Parameters
        ----------
        modelSpecification: ExtendedRoadRunner/str
            roadrunner model or antimony model
        observedData: NamedTimeseries/str
            str: path to CSV file
        parametersToFit: list-str/None
            parameters in the model that you want to fit
            if None, no parameters are fit
        selectedColumns: list-str
            species names you wish use to fit the model
            default: all columns in observedData
        parameterLowerBound: float
            lower bound for the fitting parameters
        parameterUpperBound: float
            upper bound for the fitting parameters
        parameterDct: dict
            key: parameter name
            value: triple - (lowerVange, startingValue, upperRange)
        fittedDataTransformDct: dict
            key: column in selectedColumns
            value: function of the data in selectedColumns;
                   input: NamedTimeseries
                   output: array for the values of the column
        logger: Logger
        fitterMethods: str/list-str/list-OptimizerMethod
            method used for minimization in fitModel
        numFitRepeat: int
            number of times fitting is repeated for a method
        bootstrapMethods: str/list-str/list-OptimizerMethod
            method used for minimization in bootstrap
        numIteration: number of bootstrap iterations
        reportInterval: number of iterations between progress reports
        maxProcess: Maximum number of processes to use. Default: numCPU
        serializePath: Where to serialize the fitter after bootstrap

        Usage
        -----
        parameterDct = {
            "k1": (1, 5, 10),  # name of parameter: low value, initial, high
            "k2": (2, 3, 6)}
        ftter = ModelFitter(roadrunnerModel, "observed.csv",
            parameterDct=parameterDct)
        fitter.fitModel()  # Do the fit
        fitter.bootstrap()  # Estimate parameter variance with bootstrap
        """
        if modelSpecification is not None:
            # Not the default constructor
            self._loggerPrefix = _loggerPrefix
            self.modelSpecification = modelSpecification
            self.parametersToFit = parametersToFit
            self.lowerBound = parameterLowerBound
            self.upperBound = parameterUpperBound
            self.bootstrapKwargs = dict(
                  numIteration=numIteration,
                  reportInterval=reportInterval,
                  maxProcess=maxProcess,
                  serializePath=serializePath,
                  )
            self.parameterDct = ModelFitterCore._updateParameterDct(parameterDct)
            self._numFitRepeat = numFitRepeat
            if self.parametersToFit is None:
                self.parametersToFit = list(self.parameterDct.keys())
            self.observedTS = observedData
            if self.observedTS is not None:
                self.observedTS = mkNamedTimeseries(observedData)
            #
            self.fittedDataTransformDct = fittedDataTransformDct
            #
            if (selectedColumns is None) and (self.observedTS is not None):
                selectedColumns = self.observedTS.colnames
            self.selectedColumns = selectedColumns
            if self.observedTS is not None:
                self._observedArr = self.observedTS[self.selectedColumns].flatten()
            else:
                self._observedArr = None
            # Other internal state
            self._fitterMethods = self._makeMethods(fitterMethods,
                  METHOD_FITTER_DEFAULTS)
            self._bootstrapMethods = self._makeMethods(bootstrapMethods,
                  METHOD_BOOTSTRAP_DEFAULTS)
            if isinstance(self._bootstrapMethods, str):
                self._bootstrapMethods = [self._bootstrapMethods]
            self._isPlot = isPlot
            self._plotter = tp.TimeseriesPlotter(isPlot=self._isPlot)
            self._plotFittedTS = None  # Timeseries that is plotted
            self.logger = logger
            # The following are calculated during fitting
            self.roadrunnerModel = None
            self.minimizer = None  # lmfit.minimizer
            self.minimizerResult = None  # Results of minimization
            self.params = None  # params property in lmfit.minimizer
            self.fittedTS = self.observedTS.copy(isInitialize=True)  # Initialize
            self.residualsTS = None  # Residuals for selectedColumns
            self.bootstrapResult = None  # Result from bootstrapping
            # Validation checks
            self._validateFittedDataTransformDct()
            self._bestParameters = _BestParameters(rssq=None, params=None)
        else:
            pass

    def _makeMethods(self, methods, default):
        """
        Creates a method dictionary.

        Parameters
        ----------
        methods: str/list-str/dict
            method used for minimization in fitModel
            dict: key-method, value-optional parameters

        Returns
        -------
        list-OptimizerMethod
            key: method name
            value: dict of optional parameters
        """
        if methods is None:
            methods = default
        if isinstance(methods, str):
            if methods == METHOD_BOTH:
                methods = METHOD_FITTER_DEFAULTS
            else:
                methods = [methods]
        if isinstance(methods, list):
            if isinstance(methods[0], str):
                results = [ModelFitterCore.OptimizerMethod(method=m, kwargs={})
                      for m in methods]
            else:
                results = methods
        else:
            raise RuntimeError("Must be a list")
        trues = [isinstance(m, ModelFitterCore.OptimizerMethod) for m in results]
        if not all(trues):
            raise ValueError("Invalid methods: %s" % str(methods))
        return results


    @classmethod
    def mkParameters(cls, parameterDct:dict=None,
          parametersToFit:list=None,
          logger:Logger=Logger(),
          lowerBound:float=PARAMETER_LOWER_BOUND,
          upperBound:float=PARAMETER_UPPER_BOUND)->lmfit.Parameters:
        """
        Constructs lmfit parameters based on specifications.

        Parameters
        ----------
        parameterDct: key=name, value=ParameterSpecification
        parametersToFit: list of parameters to fit
        logger: error logger
        lowerBound: lower value of range for parameters
        upperBound: upper value of range for parameters

        Returns
        -------
        lmfit.Parameters
        """
        def get(value, base_value, multiplier):
            if value is not None:
                return value
            return base_value*multiplier
        #
        if (parametersToFit is None) and (parameterDct is None):
            raise RuntimeError("Must specify one of these parameters.")
        if parameterDct is None:
            parameterDct = {}
        if parametersToFit is None:
            parametersToFit = parameterDct.keys()
        if logger is None:
            logger = logger()
        params = lmfit.Parameters()
        for parameterName in parametersToFit:
            if parameterName in parameterDct.keys():
                specification = parameterDct[parameterName]
                value = get(specification.value, specification.value, 1.0)
                if value > 0:
                    lower_factor = LOWER_PARAMETER_MULT
                    upper_factor = UPPER_PARAMETER_MULT
                else:
                    upper_factor = UPPER_PARAMETER_MULT
                    lower_factor = LOWER_PARAMETER_MULT
                lower = get(specification.lower, specification.value,
                      lower_factor)
                upper = get(specification.upper, specification.value,
                      upper_factor)
                if np.isclose(lower - upper, 0):
                    upper = 0.0001
                try:
                    params.add(parameterName, value=value, min=lower, max=upper)
                except Exception as err:
                    msg = "modelFitterCore/mkParameters parameterName %s" \
                          % parameterName
                    logger.error(msg, err)
            else:
                value = np.mean([lowerBound, upperBound])
                params.add(parameterName, value=value,
                      min=lowerBound, max=upperBound)
        return params

    @classmethod
    def initializeRoadrunnerModel(cls, modelSpecification):
        """
        Sets self.roadrunnerModel.

        Parameters
        ----------
        modelSpecification: ExtendedRoadRunner/str

        Returns
        -------
        ExtendedRoadRunner
        """
        if isinstance(modelSpecification,
              te.roadrunner.extended_roadrunner.ExtendedRoadRunner):
            roadrunnerModel = modelSpecification
        elif isinstance(modelSpecification, str):
            roadrunnerModel = te.loada(modelSpecification)
        else:
            msg = 'Invalid model.'
            msg = msg + "\nA model must either be a Roadrunner model "
            msg = msg + "an Antimony model."
            raise ValueError(msg)
        return roadrunnerModel

    @classmethod
    def setupModel(cls, roadrunner, parameters, logger=Logger()):
        """
        Sets up the model for use based on the parameter parameters

        Parameters
        ----------
        roadrunner: ExtendedRoadRunner
        parameters: lmfit.Parameters
        logger Logger
        """
        pp = parameters.valuesdict()
        for parameter in pp.keys():
            try:
                roadrunner.model[parameter] = pp[parameter]
            except Exception as err:
                msg = "_modelFitterCore.setupModel: Could not set value for %s"  \
                      % parameter
                logger.error(msg, err)

    @classmethod
    def runSimulation(cls, parameters=None,
          roadrunner=None,
          startTime=0,
          endTime=5,
          numPoint=30,
          selectedColumns=None,
          returnDataFrame=True,
          _logger=Logger(),
          _loggerPrefix="",
          ):
        """
        Runs a simulation. Defaults to parameter values in the simulation.

        Parameters
       ----------
        roadrunner: ExtendedRoadRunner/str
            Roadrunner model
        parameters: lmfit.Parameters
            lmfit parameters
        startTime: float
            start time for the simulation
        endTime: float
            end time for the simulation
        numPoint: int
            number of points in the simulation
        selectedColumns: list-str
            output columns in simulation
        returnDataFrame: bool
            return a DataFrame
        _logger: Logger
        _loggerPrefix: str


        Return
        ------
        NamedTimeseries (or None if fail to converge)
        """
        if isinstance(roadrunner, str):
            roadrunner = cls.initializeRoadrunnerModel(roadrunner)
        else:
            roadrunner.reset()
        if parameters is not None:
            # Parameters have been specified
            cls.setupModel(roadrunner, parameters, logger=_logger)
        # Do the simulation
        if selectedColumns is not None:
            newSelectedColumns = list(selectedColumns)
            if TIME not in newSelectedColumns:
                newSelectedColumns.insert(0, TIME)
            try:
                data = roadrunner.simulate(startTime, endTime, numPoint,
                      newSelectedColumns)
            except Exception as err:
                _logger.error("Roadrunner exception: ", err)
                data = None
        else:
            try:
                data = roadrunner.simulate(startTime, endTime, numPoint)
            except Exception as err:
                _logger.exception("Roadrunner exception: %s", err)
                data = None
        if data is None:
            return data
        fittedTS = NamedTimeseries(namedArray=data)
        if returnDataFrame:
            result = fittedTS.to_dataframe()
        else:
            result = fittedTS
        return result

    @classmethod
    def rpConstruct(cls):
        """
        Overrides rpickler.rpConstruct to create a method that
        constructs an instance without arguments.

        Returns
        -------
        Instance of cls
        """
        return cls(None, None, None)

    def rpRevise(self):
        """
        Overrides rpickler.
        """
        if "logger" not in self.__dict__.keys():
            self.logger = Logger()

    def _validateFittedDataTransformDct(self):
        if self.fittedDataTransformDct is not None:
            keySet = set(self.fittedDataTransformDct.keys())
            selectedColumnsSet = self.selectedColumns
            if (keySet is not None) and (selectedColumnsSet is not None):
                excess = set(keySet).difference(selectedColumnsSet)
                if len(excess) > 0:
                    msg = "Columns not in selectedColumns: %s"  % str(excess)
                    raise ValueError(msg)

    def _transformFittedTS(self, data):
        """
        Updates the fittedTS taking into account required transformations.

        Parameters
        ----------
        data: np.ndarray

        Results
        ----------
        NamedTimeseries
        """
        colnames = list(self.selectedColumns)
        colnames.insert(0, TIME)
        fittedTS = NamedTimeseries(array=data[:, :], colnames=colnames)
        if self.fittedDataTransformDct is not None:
            for column, func in self.fittedDataTransformDct.items():
                if func is not None:
                    fittedTS[column] = func(fittedTS)
        return fittedTS

    @staticmethod
    def _updateParameterDct(parameterDct):
        """
        Handles values that are tuples instead of ParameterSpecification.
        """
        if parameterDct is None:
            parameterDct = {}
        dct = dict(parameterDct)
        for name, value in parameterDct.items():
            if isinstance(value, tuple):
                dct[name] = ParameterSpecification(lower=value[0],
                      upper=value[1], value=value[2])
        return dct

    @staticmethod
    def addParameter(parameterDct: dict,
          name: str, lower: float, upper: float, value: float):
        """
        Adds a parameter to a list of parameters.

        Parameters
        ----------
        parameterDct: parameter dictionary to agument
        name: parameter name
        lower: lower range of parameter value
        upper: upper range of parameter value
        value: initial value

        Returns
        -------
        dict
        """
        parameterDct[name] = ParameterSpecification(
              lower=lower, upper=upper, value=value)

    def _adjustNames(self, antimonyModel:str, observedTS:NamedTimeseries)  \
          ->typing.Tuple[NamedTimeseries, list]:
        """
        Antimony exports can change the names of floating species
        by adding a "_" at the end. Check for this and adjust
        the names in observedTS.

        Return
        ------
        NamedTimeseries: newObservedTS
        list: newSelectedColumns
        """
        rr = te.loada(antimonyModel)
        dataNames = rr.simulate().colnames
        names = ["[%s]" % n for n in observedTS.colnames]
        missingNames = [n[1:-1] for n in set(names).difference(dataNames)]
        newSelectedColumns = list(self.selectedColumns)
        if len(missingNames) > 0:
            newObservedTS = observedTS.copy()
            self.logger.exception("Missing names in antimony export: %s"
                  % str(missingNames))
            for name in observedTS.colnames:
                missingName = "%s_" % name
                if name in missingNames:
                    newObservedTS = newObservedTS.rename(name, missingName)
                    newSelectedColumns.remove(name)
                    newSelectedColumns.append(missingName)
        else:
            newObservedTS = observedTS
        return newObservedTS, newSelectedColumns

    def copy(self, isKeepLogger=False):
        """
        Creates a copy of the model fitter.
        Preserves the user-specified settings and the results
        of bootstrapping.
        """
        if not isinstance(self.modelSpecification, str):
            try:
                modelSpecification = self.modelSpecification.getAntimony()
            except Exception as err:
                self.logger.error("Problem wth conversion to Antimony. Details:",
                      err)
                raise ValueError("Cannot proceed.")
            observedTS, selectedColumns = self._adjustNames(
                  modelSpecification, self.observedTS)
        else:
            modelSpecification = self.modelSpecification
            observedTS = self.observedTS.copy()
            selectedColumns = self.selectedColumns
        #
        if isKeepLogger:
            logger = self.logger
        elif self.logger is not None:
            logger = self.logger.copy()
        else:
            logger = None
        newModelFitter = self.__class__(
              copy.deepcopy(modelSpecification),
              observedTS,
              copy.deepcopy(self.parametersToFit),
              selectedColumns=selectedColumns,
              fitterMethods=self._fitterMethods,
              bootstrapMethods=self._bootstrapMethods,
              parameterLowerBound=self.lowerBound,
              parameterUpperBound=self.upperBound,
              parameterDct=copy.deepcopy(self.parameterDct),
              fittedDataTransformDct=copy.deepcopy(self.fittedDataTransformDct),
              logger=logger,
              isPlot=self._isPlot)
        if self.bootstrapResult is not None:
            newModelFitter.bootstrapResult = self.bootstrapResult.copy()
            newModelFitter.params = newModelFitter.bootstrapResult.params
        else:
            newModelFitter.bootstrapResult = None
            newModelFitter.params = self.params
        return newModelFitter

    def initializeRoadRunnerModel(self):
        """
        Sets self.roadrunnerModel.
        """
        self.roadrunnerModel = ModelFitterCore.initializeRoadrunnerModel(
              self.modelSpecification)

    def getDefaultParameterValues(self):
        """
        Obtain the original values of parameters.

        Returns
        -------
        dict:
            key: parameter name
            value: value of parameter
        """
        dct = {}
        self.initializeRoadRunnerModel()
        self.roadrunnerModel.reset()
        for parameterName in self.parametersToFit:
            dct[parameterName] = self.roadrunnerModel.model[parameterName]
        return dct

    def simulate(self, params=None, startTime=None, endTime=None, numPoint=None):
        """
        Runs a simulation. Defaults to parameter values in the simulation.

        Parameters
       ----------
        params: lmfit.Parameters
        startTime: float
        endTime: float
        numPoint: int

        Return
        ------
        NamedTimeseries
        """
        def setValue(default, parameter):
            # Sets to default if parameter unspecified
            if parameter is None:
                return default
            return parameter
        #
        startTime = setValue(self.observedTS.start, startTime)
        endTime = setValue(self.observedTS.end, endTime)
        numPoint = setValue(len(self.observedTS), numPoint)
        #
        if self.roadrunnerModel is None:
            self.initializeRoadRunnerModel()
        #
        return ModelFitterCore.runSimulation(parameters=params,
              roadrunner=self.roadrunnerModel,
              startTime=startTime,
              endTime=endTime,
              numPoint=numPoint,
              selectedColumns=self.selectedColumns,
              _logger=self.logger,
              _loggerPrefix=self._loggerPrefix,
              returnDataFrame=False)

    def updateFittedAndResiduals(self, **kwargs)->np.ndarray:
        """
        Updates values of self.fittedTS and self.residualsTS
        based on self.params.

        Parameters
        ----------
        kwargs: dict
            arguments for simulation

        Instance Variables Updated
        --------------------------
        self.fittedTS
        self.residualsTS

        Returns
        -------
        1-d ndarray of residuals
        """
        self.fittedTS = self.simulate(**kwargs)  # Updates self.fittedTS
        residualsArr = self._residuals(self.params)
        numRow = len(self.fittedTS)
        numCol = len(residualsArr)//numRow
        residualsArr = np.reshape(residualsArr, (numRow, numCol))
        cols = self.selectedColumns
        if self.residualsTS is None:
            self.residualsTS = self.observedTS.subsetColumns(cols)
        self.residualsTS[cols] = residualsArr

    def _residuals(self, params)->np.ndarray:
        """
        Compute the residuals between objective and experimental data
        Handle nan values in observedTS. This internal-only method
        is implemented to maximize efficieency.

        Parameters
        ----------
        kwargs: dict
            arguments for simulation

        Returns
        -------
        1-d ndarray of residuals
        """
        data = ModelFitterCore.runSimulation(parameters=params,
              roadrunner=self.roadrunnerModel,
              startTime=self.observedTS.start,
              endTime=self.observedTS.end,
              numPoint=len(self.observedTS),
              selectedColumns=self.selectedColumns,
              _logger=self.logger,
              _loggerPrefix=self._loggerPrefix,
              returnDataFrame=False)
        if data is None:
            residualsArr = np.repeat(LARGE_RESIDUAL, len(self._observedArr))
        else:
            residualsArr = self._observedArr - data.flatten()
            residualsArr = np.nan_to_num(residualsArr)
        rssq = sum(residualsArr**2)
        if (self._bestParameters.rssq is None)  \
              or (rssq < self._bestParameters.rssq):
            self._bestParameters = _BestParameters(
                  params=params.copy(), rssq=rssq)
        return residualsArr

    def fitModel(self, params:lmfit.Parameters=None, max_nfev=100):
        """
        Fits the model by adjusting values of parameters based on
        differences between simulated and provided values of
        floating species.

        Parameters
        ----------
        params: starting values of parameters

        Example
        -------
        f.fitModel()
        """
        ParameterDescriptor = collections.namedtuple("ParameterDescriptor",
              "params method rssq kwargs minimizer minimizerResult")
        MAX_NFEV = "max_nfev"
        block = Logger.join(self._loggerPrefix, "fitModel")
        guid = self.logger.startBlock(block)
        self.initializeRoadRunnerModel()
        self.params = None
        if self.parametersToFit is not None:
            if params is None:
                params = self.mkParams()
            # Fit the model to the data using one or more methods.
            # Choose the result with the lowest residual standard deviation
            paramResults = []
            lastExcp = None
            for idx, optimizerMethod in enumerate(self._fitterMethods):
                method = optimizerMethod.method
                kwargs = optimizerMethod.kwargs
                if MAX_NFEV not in kwargs:
                    kwargs[MAX_NFEV] = max_nfev
                for _ in range(self._numFitRepeat):
                    self._bestParameters = _BestParameters(params=None, rssq=None)
                    minimizer = lmfit.Minimizer(self._residuals, params)
                    try:
                        minimizerResult = minimizer.minimize(
                              method=method, **kwargs)
                    except Exception as excp:
                        lastExcp = excp
                        msg = "Error minimizing for method: %s" % method
                        self.logger.error(msg, excp)
                        continue
                    params = self._bestParameters.params.copy()
                    rssq = np.sum(self._residuals(params)**2)
                    if len(paramResults) > idx:
                        if rssq >= paramResults[idx].rssq:
                            continue
                    parameterDescriptor = ParameterDescriptor(
                          params=params,
                          method=method,
                          rssq=rssq,
                          kwargs=dict(kwargs),
                          minimizer=minimizer,
                          minimizerResult=minimizerResult,
                          )
                    paramResults.append(parameterDescriptor)
            if len(paramResults) == 0:
                msg = "*** Minimizer failed for this model and data."
                self.logger.error(msg, lastExcp)
            else:
                # Select the result that has the smallest residuals
                sortedMethods = sorted(paramResults, key=lambda r: r.rssq)
                bestMethod = sortedMethods[0]
                self.params = bestMethod.params
                self.minimizer= bestMethod.minimizer
                self.minimizerResult = bestMethod.minimizerResult
        # Ensure that residualsTS and fittedTS match the parameters
        self.updateFittedAndResiduals(params=self.params)
        self.logger.endBlock(guid)

    def getFittedModel(self):
        """
        Provides the roadrunner model with fitted parameters

        Returns
        -------
        ExtendedRoadrunner
        """
        self._checkFit()
        self.roadrunnerModel.reset()
        self._setupModel(self.params)
        return self.roadrunnerModel

    def _setupModel(self, parameters):
        """
        Sets up the model for use based on the parameter parameters

        Parameters
        ----------
        parameters: lmfit.Parameters

        """
        ModelFitterCore.setupModel(self.roadrunnerModel, parameters, logger=self.logger)

    def mkParams(self, parameterDct:dict=None)->lmfit.Parameters:
        """
        Constructs lmfit parameters based on specifications.

        Parameters
        ----------
        parameterDct: key=name, value=ParameterSpecification

        Returns
        -------
        lmfit.Parameters
        """
        if parameterDct is None:
            parameterDct = self.parameterDct
        return ModelFitterCore.mkParameters(parameterDct,
              parametersToFit=self.parametersToFit,
              logger=self.logger,
              lowerBound=self.lowerBound,
              upperBound=self.upperBound)

    def _checkFit(self):
        if self.params is None:
            raise ValueError("Must use fitModel before using this method.")

    def serialize(self, path):
        """
        Serialize the model to a path.

        Parameters
        ----------
        path: str
            File path
        """
        newModelFitter = self.copy()
        with open(path, "wb") as fd:
            rpickle.dump(newModelFitter, fd)

    @classmethod
    def deserialize(cls, path):
        """
        Deserialize the model from a path.

        Parameters
        ----------
        path: str
            File path

        Return
        ------
        ModelFitter
            Model is initialized.
        """
        with open(path, "rb") as fd:
            fitter = rpickle.load(fd)
        fitter.initializeRoadRunnerModel()
        return fitter



# -*- coding: utf-8 -*-
"""
 Created on August 18, 2020

@author: joseph-hellerstein

Bootstrapping. Provides a single external: bootstrap().
Several considerations are made;
  1. Extensibility of the bootstrap results by returning
     instances of BootstrapResult.
  2. Multiprocessing. This requires a top-level (static)
     method. Arguments are packaged in _Arguments.
"""

from SBstoat.timeseriesStatistic import TimeseriesStatistic
from SBstoat._bootstrapResult import BootstrapResult
from SBstoat.observationSynthesizer import  \
      ObservationSynthesizerRandomizedResiduals
from SBstoat import _modelFitterCore as mfc
from SBstoat import _helpers
from SBstoat.logs import Logger

import multiprocessing
import numpy as np
import sys
import typing

MAX_CHISQ_MULT = 5
PERCENTILES = [2.5, 97.55]  # Percentile for confidence limits
IS_REPORT = True
ITERATION_MULTIPLIER = 10  # Multiplier to calculate max bootsrap iterations
ITERATION_PER_PROCESS = 20  # Numer of iterations handled by a process
MAX_TRIES = 10  # Maximum number of tries to fit
MAX_ITERATION_TIME = 10.0


###############  HELPER CLASSES ###############
class _Arguments():
    """ Arguments passed to _runBootstrap. """

    def __init__(self, fitter, numProcess:int, processIdx:int,
                 numIteration:int=10,
                 reportInterval:int=-1,
                 synthesizerClass=  \
                 ObservationSynthesizerRandomizedResiduals,
                 _loggerPrefix="",
                 **kwargs: dict):
        # Same the antimony model, not roadrunner bcause of Pickle
        self.fitter = fitter.copy(isKeepLogger=True)
        self.numIteration  = numIteration
        self.reportInterval  = reportInterval
        self.numProcess = numProcess
        self.processIdx = processIdx
        self.synthesizerClass = synthesizerClass
        self._loggerPrefix = _loggerPrefix
        self.kwargs = kwargs


################# FUNCTIONS ####################
def _runBootstrap(arguments:_Arguments, queue=None)->BootstrapResult:
    """
    Executes bootstrapping.

    Parameters
    ----------
    arguments: inputs to bootstrap
    queue: multiprocessing.Queue

    Notes
    -----
    1. Only the first process generates progress reports.
    2. Uses METHOD_LEASTSQ for fitModel iterations.
    """
    fitter = arguments.fitter
    logger = fitter.logger
    mainBlock = Logger.join(arguments._loggerPrefix, "_runBootstrap")
    mainGuid = logger.startBlock(mainBlock)
    # Unapack arguments
    isSuccess = False
    lastErr = ""
    # Do an initial fit
    for _ in range(MAX_TRIES):
        try:
            fitter.fitModel()  # Initialize model
            isSuccess = True
            break
        except Exception as err:
            lastErr = err
    # Set up logging for this process
    fd = logger.getFileDescriptor()
    processIdx = arguments.processIdx
    if fd is not None:
        sys.stderr = logger.getFileDescriptor()
        sys.stdout = logger.getFileDescriptor()
    iterationGuid = None
    if not isSuccess:
        msg = "Process %d/modelFitterBootstrip/_runBootstrap" % processIdx
        logger.error(msg,  lastErr)
        fittedStatistic = TimeseriesStatistic(fitter.observedTS, percentiles=[])
        bootstrapResult = BootstrapResult(fitter, 0, {},
              fittedStatistic)
    else:
        numIteration = arguments.numIteration
        reportInterval = arguments.reportInterval
        processingRate = min(arguments.numProcess,
                             multiprocessing.cpu_count())
        cols = fitter.selectedColumns
        synthesizer = arguments.synthesizerClass(
              observedTS=fitter.observedTS.subsetColumns(cols),
              fittedTS=fitter.fittedTS.subsetColumns(cols),
              **arguments.kwargs)
        # Initialize
        parameterDct = {p: [] for p in fitter.parametersToFit}
        numSuccessIteration = 0
        lastReport = 0
        if fitter.minimizerResult is None:
            fitter.fitModel()
        baseChisq = fitter.minimizerResult.redchi
        # Do the bootstrap iterations
        bootstrapError = 0
        iterationBlock = Logger.join(mainBlock, "Iteration")
        for iteration in range(numIteration*ITERATION_MULTIPLIER):
            if iterationGuid is not None:
                logger.endBlock(iterationGuid)
            iterationGuid = logger.startBlock(iterationBlock)
            newObservedTS = synthesizer.calculate()
            fittingSetupBlock = Logger.join(iterationBlock, "fittingSetup")
            fittingSetupGuid = logger.startBlock(fittingSetupBlock)
            newFitter = ModelFitterBootstrap(fitter.roadrunnerModel,
                  newObservedTS,
                  fitter.parametersToFit,
                  selectedColumns=fitter.selectedColumns,
                  # Use bootstrap methods for fitting
                  fitterMethods=fitter._bootstrapMethods,
                  parameterLowerBound=fitter.lowerBound,
                  parameterUpperBound=fitter.upperBound,
                  fittedDataTransformDct=fitter.fittedDataTransformDct,
                  logger=logger,
                  _loggerPrefix=iterationBlock,
                  isPlot=fitter._isPlot)
            fittedStatistic = TimeseriesStatistic(newFitter.observedTS,
                  percentiles=[])
            logger.endBlock(fittingSetupGuid)
            try:
                if (iteration > 0) and (iteration != lastReport)  \
                        and (processIdx == 0):
                    totalSuccessIteration = numSuccessIteration*processingRate
                    totalIteration = iteration*processingRate
                    if totalIteration % reportInterval == 0:
                        msg = "Bootstrap completed %d total iterations "
                        msg += "with %d successes."
                        msg = msg % (totalIteration, totalSuccessIteration)
                        fitter.logger.status(msg)
                        lastReport = numSuccessIteration
                if numSuccessIteration >= numIteration:
                    # Performed the iterations
                    break
                tryBlock = Logger.join(iterationBlock, "try")
                tryGuid = logger.startBlock(tryBlock)
                try:
                    tryFitterBlock = Logger.join(tryBlock, "Fitter")
                    tryFitterGuid = logger.startBlock(tryFitterBlock)
                    newFitter.fitModel(params=fitter.params)
                    logger.endBlock(tryFitterGuid)
                except Exception as err:
                    # Problem with the fit. Don't numSuccessIteration it.
                    msg = "Process %d/modelFitterBootstrap" % processIdx
                    msg += " Fit failed on iteration %d."  % iteration
                    fitter.logger.error(msg, err)
                    logger.endBlock(tryGuid)
                    continue
                if newFitter.minimizerResult.redchi > MAX_CHISQ_MULT*baseChisq:
                    if IS_REPORT:
                        msg = "Process %d: Fit has high chisq: %2.2f on iteration %d."
                        fitter.logger.exception(msg % (processIdx,
                              newFitter.minimizerResult.redchi, iteration))
                    logger.endBlock(tryGuid)
                    continue
                if newFitter.params is None:
                    continue
                numSuccessIteration += 1
                dct = newFitter.params.valuesdict()
                [parameterDct[p].append(dct[p]) for p in fitter.parametersToFit]
                cols = newFitter.fittedTS.colnames
                fittedStatistic.accumulate(newFitter.fittedTS)
                newFitter.observedTS = synthesizer.calculate()
                logger.endBlock(tryGuid)
            except Exception as err:
                msg = "Process %d/modelFitterBootstrap" % processIdx
                msg += " Error on iteration %d."  % iteration
                fitter.logger.error(msg, err)
                bootstrapError += 1
        fitter.logger.status("Process %d: completed bootstrap." % (processIdx + 1))
        bootstrapResult = BootstrapResult(fitter, numSuccessIteration, parameterDct,
              fittedStatistic, bootstrapError=bootstrapError)
    if iterationGuid is not None:
        logger.endBlock(iterationGuid)
    logger.endBlock(mainGuid)
    if fd is not None:
        if not fd.closed:
            fd.close()
    if queue is None:
        return bootstrapResult
    queue.put(bootstrapResult)


##################### CLASSES #########################
class ModelFitterBootstrap(mfc.ModelFitterCore):

    def bootstrap(self,
          # The following must be kept in sync with ModelFitterCore.__init__
          numIteration:int=None,
          reportInterval:int=None,
          synthesizerClass=ObservationSynthesizerRandomizedResiduals,
          maxProcess:int=None,
          serializePath:str=None,
          **kwargs: dict):
        """
        Constructs a bootstrap estimate of parameter values.

        Parameters
        ----------
        numIteration: number of bootstrap iterations
        reportInterval: number of iterations between progress reports
        synthesizerClass: object that synthesizes new observations
            Must subclass ObservationSynthesizer
        maxProcess: Maximum number of processes to use. Default: numCPU
        serializePath: Where to serialize the fitter after bootstrap
        kwargs: arguments passed to ObservationSynthesizer

        Example
        -------
            f.bootstrap()
            f.getParameterMeans()  # Mean values
            f.getParameterStds()  # standard deviations

        Notes
            1. Arguments can be overriden by the constructor using
               the keyword argument bootstrapKwargs.
        ----
        """
        def get(name, value):
            if value is not None:
                return value
            # Handle arguments specified in constructor
            if name in self.bootstrapKwargs:
                if self.bootstrapKwargs[name] is not None:
                    return self.bootstrapKwargs[name]
            # None specified
            return None
        # Handle overrides of arguments specified in constructor
        numIteration = get("numIteration", numIteration)
        reportInterval = get("reportInterval", reportInterval)
        synthesizerClass = get("synthesizerClass", synthesizerClass)
        maxProcess = get("maxProcess", maxProcess)
        serializePath = get("serializePath", serializePath)
        # Other initializations
        if maxProcess is None:
            maxProcess = multiprocessing.cpu_count()
        if self.minimizerResult is None:
            self.fitModel()
        # Run processes
        numProcess = max(int(numIteration/ITERATION_PER_PROCESS), 1)
        numProcess = min(numProcess, maxProcess)
        numProcessIteration = int(np.ceil(numIteration/numProcess))
        msg = "Running bootstrap for %d successful iterations " % numIteration
        msg += "with %d processes." % numProcess
        self.logger.activity(msg)
        # Run separate processes for each bootstrap
        processes = []
        queue = multiprocessing.Queue()
        results = []
        # Set to False for debug so not doing multiple processes
        if True:
            args_list = [_Arguments(self, numProcess, i,
                  numIteration=numProcessIteration,
                  reportInterval=reportInterval,
                  synthesizerClass=synthesizerClass,
                  _loggerPrefix="bootstrap",
                  **kwargs) for i in range(numProcess)]
            for args in args_list:
                p = multiprocessing.Process(target=_runBootstrap,
                      args=(args, queue,))
                p.start()
                processes.append(p)
            timeout = MAX_ITERATION_TIME*numProcessIteration
            try:
                # Get rid of possible zombies
                for _ in range(len(processes)):
                    results.append(queue.get(timeout=timeout))
                # Get rid of possible zombies
                for process in processes:
                    process.terminate()
            except Exception as err:
                msg = "modelFitterBootstrap/Error in process management"
                self.logger.error(msg, err)
            finally:
                pass
        else:
            # Keep to debug _runBootstrap single threaded
            thisNumProcess = 1
            thisProcessIdx = 0
            args = _Arguments(self, thisNumProcess, thisProcessIdx,
                  numIteration=numIteration,
                  reportInterval=reportInterval,
                  synthesizerClass=synthesizerClass,
                  _loggerPrefix="bootstrap",
                  **kwargs)
            results = [_runBootstrap(args)]
        if len(results) == 0:
            msg = "modelFitterBootstrap/timeout in solving model."
            msg = "\nConsider increasing per timeout."
            msg = "\nCurent value: %f" % MAX_ITERATION_TIME
            self.logger.result(msg)
        else:
            self.bootstrapResult = BootstrapResult.merge(results)
            # Update the logger in place
            _ = _helpers.copyObject(self.bootstrapResult.fitter.logger,
                  self.logger)
            if self.bootstrapResult.fittedStatistic is not None:
                self.bootstrapResult.fittedStatistic.calculate()
            self.logger.result("%d bootstrap estimates of parameters."
                  % self.bootstrapResult.numSimulation)
            if serializePath is not None:
                self.serialize(serializePath)

    def getParameterMeans(self)->typing.List[float]:
        """
        Returns a list of values mean values of parameters from bootstrap.

        Return
        ------
        NamedTimeseries

        Example
        -------
              f.getParameterMeans()
        """
        if self._checkBootstrap(isError=False):
            return self.bootstrapResult.parameterMeanDct.values()
        return [self.params[p].value for p in self.parametersToFit]

    def getParameterStds(self)->typing.List[float]:
        """
        Returns a list of values std values of parameters from bootstrap.

        Return
        ------
        NamedTimeseries

        Example
        -------
              f.getParameterStds()
        """
        if self._checkBootstrap(isError=False):
            return list(self.bootstrapResult.parameterStdDct.values())
        raise ValueError("***Must run bootstrap to get parameter stds.")

    def _checkBootstrap(self, isError:bool=True)->bool:
        """
        Verifies that bootstrap has been done.
        """
        self._checkFit()
        if self.bootstrapResult is None:
            if isError:
                raise ValueError("Must use bootstrap first.")
            return False
        return True



"""
Runs TestHarness for BioModels. Creates:
    *.png figure that plots relative errors
    *.pcl file with data collected from run
    *.log file with information about run

 Common usage:

  # Access information about command arguments
  python SBstoat/mainTestHarness.py --help

  # Process the BioModels 1-800, creating a new log file and data file
  python SBstoat/mainTestHarness.py --firstModel 1 --numModel 800

  # Process the BioModels 1-800, using the existing log and data files.
  python SBstoat/mainTestHarness.py --firstModel 1 --numModel 800 --useExistingData --useExistingLog

  # Create a plot from the existing data file
  python SBstoat/mainTestHarness.py --plot

  # Run analysis

@author: joseph-hellerstein
"""

from SBstoat._testHarness import TestHarness
from SBstoat.logs import Logger

import argparse
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import os
import pickle

# Handle problem with module load
try:
    matplotlib.use('TkAgg')
except ImportError:
    pass

IGNORE_TEST = True
IS_PLOT = True
DIR = os.path.dirname(os.path.abspath(__file__))
DATA_DIR = os.path.join(os.path.dirname(DIR), "biomodels")
PATH_PAT = os.path.join(DATA_DIR, "BIOMD0000000%03d.xml")
LOG_PATH = os.path.join(DIR, "mainTestHarness.log")
FIG_PATH = os.path.join(DIR, "mainTestHarness.png")
FIRST_MODEL = 210
NUM_MODEL = 2
PCL_FILE = "mainTestHarness.pcl"
FIT_MODEL = "fitModel"
BOOTSTRAP = "bootstrap"
NUM_NOERROR = "num_noerror"
NUM_MODEL = "num_model"
LOGGER = "logger"
# Context variables that are saved. Uses the following naming convention:
#  ends in "s" is a list: initialized to []
#  ends in "Dct" is a dict: initialized to {}
#  ends in "Path" is file path: initialized to None
#  begins with "is" is a bool: initialized to False
#  otherwise: int: initialized to 0
CONTEXT =  [ "firstModel", "numModel", "numNoError", "fitModelRelerrors",
      "bootstrapRelerrors", "processedModels", "nonErroredModels", "erroredModels",
      "modelParameterDct", "pclPath", "figPath", "isPlot", "reportInterval",
      "kwargDct"
      ]


############### FUNCTIONS ##################
def str2Bool(v):
    if isinstance(v, bool):
       return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

def remove(ffile):
    if os.path.isfile(ffile):
        os.remove(ffile)


############### CLASSES ##################
class Runner(object):

    """Runs tests on biomodels."""

    def __init__(self, firstModel:int=210, numModel:int=2,
          pclPath=PCL_FILE, figPath=FIG_PATH,
          useExistingData:bool=False, reportInterval:int=10,
          isPlot=IS_PLOT, **kwargDct):
        """
        Parameters
        ----------
        firstModel: first model to use
        numModel: number of models to use
        pclPath: file to which results are saved
        reportInterval: how frequently report progress
        useExistingData: use data in existing PCL file
        """
        self.useExistingData = useExistingData and os.path.isfile(pclPath)
        # Recover previously saved results if desired
        if self.useExistingData:
            self.restore(pclPath=pclPath)
        else:
            # Initialize based on type of context variable
            for name in CONTEXT:
                if name[-1:] == "s":
                    self.__setattr__(name, [])
                elif name[-3:]  == "Dct":
                    self.__setattr__(name, {})
                elif name[-4:]  == "Path":
                    self.__setattr__(name, None)
                elif name[0:2]  == "is":
                    self.__setattr__(name, False)
                else:
                    self.__setattr__(name, 0)
        # Initialize to parameters for this instantiation
        self.firstModel = firstModel
        self.numModel = numModel
        self.pclPath = pclPath
        self.figPath = figPath
        self.reportInterval = reportInterval
        self.kwargDct = kwargDct
        self.isPlot = isPlot
        self.useExistingData = useExistingData
        #
        if LOGGER in kwargDct.keys():
            self.logger = kwargDct[LOGGER]
        else:
            self.logger = Logger()
            kwargDct[LOGGER] = self.logger
        self.save()

    def _isListSame(self, list1, list2):
        diff = set(list1).symmetric_difference(list2)
        return len(diff) == 0

    def equals(self, other):
        selfKeys = list(self.__dict__.keys())
        otherKeys = list(other.__dict__.keys())
        if not self._isListSame(selfKeys, otherKeys):
            return False
        #
        for key, value in self.__dict__.items():
            if isinstance(value, list):
                isEqual = self._isListSame(value, other.__getattribute__(key))
                if not isEqual:
                    import pdb; pdb.set_trace()
                    return False
            elif any([isinstance(value, t) for t in [int, str, float, bool]]):
                if self.__getattribute__(key) != other.__getattribute__(key):
                    import pdb; pdb.set_trace()
                    return False
            else:
                pass
        #
        return True

    def run(self):
        """
        Runs the tests. Saves state after each tests.
        """
        # Processing models
        modelNums = self.firstModel + np.array(range(self.numModel))
        for modelNum in modelNums:
            if (modelNum in self.processedModels) and self.useExistingData:
                continue
            else:
                self.processedModels.append(modelNum)
                input_path = PATH_PAT % modelNum
                msg = "Model %s" % input_path
                self.logger.activity(msg)
                try:
                    harness = TestHarness(input_path, **self.kwargDct)
                    if len(harness.parametersToFit) == 0:
                        self.logger.result("No fitable parameters in model.")
                        self.save()
                        continue
                    harness.evaluate(stdResiduals=1.0,
                          fractionParameterDeviation=1.0, relError=2.0)
                except Exception as err:
                    self.erroredModels.append(modelNum)
                    self.logger.error("TestHarness failed", err)
                    self.save()
                    continue
                # Parameters for model
                self.modelParameterDct[modelNum] =  \
                      list(harness.fitModelResult.parameterRelErrorDct.keys())
                # Relative error in initial fit
                values = [v for v in
                      harness.fitModelResult.parameterRelErrorDct.values()]
                self.fitModelRelerrors.extend(values)
                # Relative error in bootstrap
                values = [v for v in
                      harness.bootstrapResult.parameterRelErrorDct.values()]
                self.bootstrapRelerrors.extend(values)
                # Count models without exceptions
                self.nonErroredModels.append(modelNum)
                self.numNoError = len(self.nonErroredModels)
                if modelNum % self.reportInterval == 0:
                    self.logger.result("Processed model %d" % modelNum)
                self.save()
        # Check for plot
        if self.isPlot:
            self.plot()

    def save(self):
        """
        Saves state. Maintain in sync with self.restore().
        """
        if self.pclPath is not None:
            data = [self.__getattribute__(n) for n in CONTEXT]
            with (open(self.pclPath, "wb")) as fd:
                pickle.dump(data, fd)

    def restore(self, pclPath=None):
        """
        Restores state. Maintain in sync with self.save().
        """
        if pclPath is None:
            pclPath = self.pclPath
        if os.path.isfile(pclPath):
            with (open(pclPath, "rb")) as fd:
                data = pickle.load(fd)
            [self.__setattr__(n, v) for n, v in zip(CONTEXT, data)]
        else:
            raise ValueError("***Restart file %s does not exist"
                  % self.pclPath)

    def plot(self):
        """
        Does all plots.
        """
        _, axes = plt.subplots(1, 2)
        maxBin1 = self._plotRelativeErrors(axes[0], self.fitModelRelerrors,
              FIT_MODEL)
        maxBin2 = self._plotRelativeErrors(axes[1], self.bootstrapRelerrors,
              BOOTSTRAP, isYLabel=False)
        maxBin = max(maxBin1, maxBin2)
        if maxBin > 0:
            axes[0].set_ylim([0, maxBin])
            axes[1].set_ylim([0, maxBin])
        #
        if len(self.processedModels) == 0:
            frac = 0.0
        else:
            frac = 1.0*self.numNoError/len(self.processedModels)
        suptitle = "Models %d-%d. Fraction non-errored: %2.3f"
        lastModel = self.firstModel + len(self.processedModels) - 1
        suptitle = suptitle % (self.firstModel, lastModel, frac)
        plt.suptitle(suptitle)
        plt.show()
        plt.savefig(self.figPath)

    def _plotRelativeErrors(self, ax, relErrors, title, isYLabel=True):
        """
        Plots histogram of relative errors.

        Parameters
        ----------
        ax: Matplotlib.axes
        relErrors: list-float
        title: str

        Returns
        -------
        float: maximum number in a bin
        """
        rr = ax.hist(relErrors)
        ax.set_title(title)
        ax.set_xlabel("relative error")
        if isYLabel:
            ax.set_ylabel("number parameters")
        ax.set_xlim([0, 1])
        return max(rr[0])


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='SBstoat tests for BioModels.')
    default = 1
    parser.add_argument('--firstModel', type=int,
        help='First BioModel to process (int); default: %d' % default,
        default=default)
    default = 0
    parser.add_argument('--numModel', type=int,
        help='Number of models to process (int); default = %d' % default,
        default=default)
    parser.add_argument('--logPath', type=str,
        help='Path for log file (str); default: %s' % LOG_PATH,
        default=LOG_PATH)
    parser.add_argument('--figPath', type=str,
        help='Path for figure (str); Default: %s' % FIG_PATH,
        default=FIG_PATH)
    parser.add_argument('--useExistingData', action='store_true',
        help="Use saved data from an previous run (flag).")
    parser.add_argument('--plot', action='store_true',
        help="Plot existing data (flag).")
    parser.add_argument('--useExistingLog', action='store_true',
        help="Append to the existing log file, if it exists (flag).")
    args = parser.parse_args()
    useExistingLog = args.plot or args.useExistingLog
    useExistingData = (args.plot and (args.numModel == 0)) or args.useExistingData
    #
    if not useExistingLog:
        remove(args.logPath)
    runner = Runner(firstModel=args.firstModel,
                    numModel=args.numModel,
                    useExistingData=useExistingData,
                    figPath=args.figPath,
                    isPlot=args.plot,
                    logger=Logger(toFile=args.logPath))
    runner.run()



"""Helper functions used in SBstoat."""


from SBstoat import _plotOptions as po

import copy
import inspect
import numpy as np
import scipy

INDENTATION = "  "
NULL_STR = ""


def updatePlotDocstring(target):
    """
    Changes the docstring of plot function to include all
    plot options.

    Parameters
    ----------
    target: class or function
    keyprhase: string searched for in docstring
    """
    # Place import here to avoid circular dependencies
    plot_options = str(po.PlotOptions())
    def updateFunctionDocstring(func):
        docstring = func.__doc__
        if not po.EXPAND_KEYPHRASE in docstring:
            msg = "Keyword not found in method: %s"  \
                  % func.__name__
            raise RuntimeError(msg)
        new_docstring =  \
              docstring.replace(
                    po.EXPAND_KEYPHRASE, plot_options)
        func.__doc__ = new_docstring
    #
    if "__call__" in dir(target):
        # Handle a function
        updateFunctionDocstring(target)
    else:
        # Update a class
        cls = target
        for name in dir(cls):
            if name[0:4] == po.PLOT:
                method = eval("cls.%s" % name)
                updateFunctionDocstring(method)


class Report():
    """Class used to generate reports."""

    def __init__(self):
        self.reportStr= NULL_STR
        self.numIndent = 0

    def indent(self, num: int):
        self.numIndent += num

    def _getIndentStr(self):
        return NULL_STR.join(np.repeat(
              INDENTATION, self.numIndent))

    def addHeader(self, title:str):
        indentStr = self._getIndentStr()
        self.reportStr+= "\n%s%s" % (indentStr, title)

    def addTerm(self, name:str, value:object):
        indentStr = self._getIndentStr()
        self.reportStr+= "\n%s%s: %s" %  \
              (indentStr, name, str(value))

    def get(self)->str:
        return self.reportStr

def calcRelError(actual:float, estimated:float, isAbsolute:bool=True):
    """
    Calculates the relative error of the estimate.

    Parameters
    ----------
    actual: actual value
    estimated: estimated values
    isAbsolute: return absolute value

    Returns
    -------
    float
    """
    if np.isclose(actual, 0):
        return np.nan
    relError = (estimated - actual) / actual
    if isAbsolute:
        relError = np.abs(relError)
    return relError

def filterOutliersFromZero(data, maxSL):
    """
    Removes values that are distant from 0 using a F-statistic criteria.
    Extreme values are iteratively removed until the F-statistic exceeds
    the significance level.

    Parameters
    ----------
    data: iterable-float
    maxSL: float
        Maximum significance level to accept a difference in variance
        A larger maxSL means more filtering since it's more likely that an
        extreme value will be filtered.

    Returns
    -------
    np.array
    """
    def calcSL(arr1, arr2):
        """
        Calculates the significance level that the variance of the first array
        is larger than the variance of the second array.

        Returns
        -------
        float
        """
        def calc(arr):
            return np.var(arr), len(arr) - 1
        #
        var1, df1 = calc(arr1)
        var2, df2 = calc(arr2)
        if var2 > 0:
            fstat = var1/var2
        else:
            fstat = 1000*var1
        sl = 1 - scipy.stats.f.cdf(fstat, df1, df2)
        return sl
    sortedData = sorted(data, key=lambda v: np.abs(v), reverse=True)
    sortedData = np.array(sortedData)
    # Construct the array without outliers
    for _ in range(len(data)):
        shortSortedData = sortedData[1:]
        sl = calcSL(sortedData, shortSortedData)
        if sl < maxSL:
            sortedData = shortSortedData
        else:
            break
    #
    return sortedData

def copyObject(oldObject, newInstance=None):
    """
    Copies the non "__" instance variables of the old object into the new instance.

    Parameters
    ----------
    oldObject: an existing object
    newInstance: updated
    """
    if newInstance is None:
        newInstance = oldObject.__class__()
    for attr in oldObject.__dict__:
        if len(attr) >= 2:
            if attr[0:2] == "__":
                continue
        value = oldObject.__getattribute__(attr)
        if "copy" in dir(value):
            newValue = value.copy()
        else:
            newValue = copy.deepcopy(value)
        try:
            newInstance.__setattr__(attr, newValue)
        except:
            continue
    return newInstance

def getKwargNames(func):
    """
    Obtains the keyword arguments in the function definition.

    Parameters
    ----------
    func: Function

    Returns
    -------
    list-str
    """
    spec = inspect.getfullargspec(func)
    argList = spec.args
    numKwarg = len(spec.defaults)
    kwargNames = argList[-numKwarg:]
    return kwargNames

def kwargs():
    """
    Decorator that provides adds properties to a function:
        defined: keyword arguments in function definition
        passed: keyword arguments/values passed
    """
    def decorator(function):
        def inner(*args, **kwwargs):
            inner.defined = getKwargNames(function)
            inner.passed = kwwargs
            inner.name = function.__qualname__
            return function(*args, **kwwargs)
        return inner
    return decorator

def validateKwargs(function):
    """
    Validates that the keywords passed to the function are a subset of the
    parameters defined.
    The function must use the @kwargs decorator.

    Parameters
    ----------
    function: function

    Raises: ValueError
    """
    missing = [p for p in function.passed.keys() for p in function.defined]
    if len(missing) > 0:
        raise ValueError(
              "The following keyword parameters do not match: %s" % str(missing))

def ppDict(dct, indent=0):
    """
    Does a pretty print of a dictionary.values.

    Parameters
    ----------
    dct: dict
    indent: int
        spaes indented

    Returns
    -------
    str
    """
    spaces = "".join([" " for _ in range(indent)])
    lines = []
    for key, value in dct.items():
        line = "%s%s:  %s" % (spaces, key, value)
        lines.append(line)
    return "\n".join(lines)



# -*- coding: utf-8 -*-
"""
 Created on August 18, 2020

@author: joseph-hellerstein

Container for the results of bootstrapping. Provides
metrics that are calculated from the results.
"""

from SBstoat.timeseriesStatistic import TimeseriesStatistic
from SBstoat.logs import Logger
from SBstoat import rpickle
from SBstoat import _helpers
from SBstoat import _modelFitterCore as mfc

import copy
import lmfit
import numpy as np
import pandas as pd

PERCENTILES = [2.5, 50, 97.55]  # Percentile for confidence limits
MIN_COUNT_PERCENTILE = 100  # Minimum number of values required to get percentiles
MIN_VALUE = 1e-3


######### CLASSES ###############
class BootstrapResult(rpickle.RPickler):

    def __init__(self, fitter, numIteration: int, parameterDct:dict,
          fittedStatistic: TimeseriesStatistic, bootstrapError=0):
        """
        Results from bootstrap

        Parameters
        ----------
        numIteration: number of successful iterations
        parameterDct: dict
            key: parameter name
            value: list of values
        fittedStatistic: statistics for fitted timeseries
        err: Error encountered
        """
        self.fitter = fitter
        self.numIteration = numIteration
        self.parameterDct = parameterDct
        self.bootstrapError = bootstrapError
        # Timeseries statistics for fits
        self.fittedStatistic = fittedStatistic
        if fitter is not None:
            self.fitter = self.fitter.copy()
            # population of parameter values
            self.parameterDct = dict(self.parameterDct)
            # list of parameters
            self.parameters = list(self.parameterDct.keys())
            # Number of simulations
            if len(self.parameters) > 0:
                self.numSimulation =  \
                      len(self.parameterDct[self.parameters[0]])
            else:
                self.numSimulation = 0
            if self.numSimulation > 1:
                # means of parameter values
                self.parameterMeanDct = {p: np.mean(parameterDct[p])
                      for p in self.parameters}
                # standard deviation of parameter values
                self.parameterStdDct = {p: np.std(parameterDct[p])
                      for p in self.parameters}
                # Confidence limits for parameter values
                self.percentileDct = {p: [] for p in self.parameterDct}
                for name, values in self.parameterDct.items():
                    if len(values) > MIN_COUNT_PERCENTILE:
                        self.percentileDct[name] = np.percentile(values, PERCENTILES)
            else:
                # means of parameter values
                self.parameterMeanDct = {p: np.nan for p in self.parameters}
                # standard deviation of parameter values
                self.parameterStdDct = {p: np.nan for p in self.parameters}
                # Confidence limits for parameter values
                self.percentileDct = {p: np.nan for p in self.parameters}
        ### PRIVATE
        # Fitting parameters from result
        self._params = None

    @classmethod
    def rpConstruct(cls):
        """
        Overrides rpickler.rpConstruct to create a method that
        constructs an instance without arguments.

        Returns
        -------
        Instance of cls
        """
        return cls(None, None, None, None)

    def copy(self):
        return copy.deepcopy(self)

    def __str__(self) -> str:
        """
        Bootstrap report.
        """
        report = _helpers.Report()
        report.addHeader("Bootstrap Report.")
        report.addTerm("Total iterations", self.numIteration)
        report.addTerm("Total simulation", self.numSimulation)
        for par in self.parameters:
            report.addHeader(par)
            report.indent(1)
            report.addTerm("mean", self.parameterMeanDct[par])
            report.addTerm("std", self.parameterStdDct[par])
            report.addTerm("%s Percentiles" % str(PERCENTILES),
                  self.percentileDct[par])
            report.indent(-1)
        return report.get()

    @property
    def params(self)->lmfit.Parameters:
        """
        Constructs parameters from bootstrap result.

        Returns
        -------
        """
        if "_params" not in self.__dict__.keys():
            self._params = None
        if self._params is None:
            self._params = lmfit.Parameters()
            for name in self.parameterMeanDct.keys():
                value = self.parameterMeanDct[name]
                if np.isclose(value, 0):
                    minValue = -MIN_VALUE
                    maxValue = MIN_VALUE
                else:
                    values = (value*0.99, value*1.01)
                    minValue = min(values)
                    maxValue = max(values)
                self._params.add(name, value=value, min=minValue, max=maxValue)
        return self._params

    def simulate(self, numSample:int=1000, numPoint:int=100)->TimeseriesStatistic:
        """
        Runs a simulation using the parameters from the bootstrap.

        Parameters
        ----------
        numSample: number of fitted parameters to sample
        numPoint: number of points in the simulation

        Returns
        -------
        TimeseriesStatistic
        """
        params_list = self._sampleParams(numSample)
        fittedTS = self.fitter.simulate(params=params_list[0], numPoint=numPoint)
        timeseriesStatistic = TimeseriesStatistic(fittedTS,
              percentiles=PERCENTILES)
        timeseriesStatistic.accumulate(fittedTS)
        # Do the simulation
        for params in params_list[1:]:
            fittedTS = self.fitter.simulate(params=params, numPoint=numPoint)
            timeseriesStatistic.accumulate(fittedTS)
        timeseriesStatistic.calculate()
        return timeseriesStatistic

    def _sampleParams(self, numSample:int):
        """
        Samples the parameters obtained in bootstrapping.

        Parameters
        ----------
        numSample: number of samples returned

        Returns
        -------
        list-lmfit.Parameters
        """
        df = pd.DataFrame(self.parameterDct)
        df_sample = df.sample(numSample, replace=True, axis=0)
        dcts = df_sample.to_dict('records')
        results = []
        for dct in dcts:
            parameterDct = {}
            for name in dct.keys():
                value = dct[name]
                parameterDct[name] = mfc.ParameterSpecification(
                      lower=value*0.9, value=value, upper=value*1.1)
            params = self.fitter.mkParams(parameterDct=parameterDct)
            results.append(params)
        return results

    @classmethod
    def merge(cls, bootstrapResults):
        """
        Combines a list of BootstrapResult.

        Parameter
        ---------
        bootstrapResults: list-BootstrapResult

        Return
        ------
        BootstrapResult
        """
        if len(bootstrapResults) == 0:
            raise ValueError("Must provide a non-empty list")
        fitter = bootstrapResults[0].fitter
        parameterDct = {p: [] for p in bootstrapResults[0].parameterDct}
        numIteration = sum([r.numIteration for r in bootstrapResults])
        bootstrapError = sum([b.bootstrapError for b in bootstrapResults])
        fittedStatistic = None
        if numIteration > 0:
            # Merge the fitter logs
            fitter.logger = Logger.merge([b.fitter.logger
                  for b in bootstrapResults])
            # Merge the statistics for fitted timeseries
            fittedStatistics = [b.fittedStatistic for b in bootstrapResults]
            fittedStatistic = TimeseriesStatistic.merge(fittedStatistics)
            # Accumulate the results
            for bootstrapResult in bootstrapResults:
                for parameter in parameterDct.keys():
                    parameterDct[parameter].extend(
                          bootstrapResult.parameterDct[parameter])
            #
        return BootstrapResult(fitter, numIteration, parameterDct,
              fittedStatistic, bootstrapError=bootstrapError)



# -*- coding: utf-8 -*-
"""
 Created on Tue Jul  7 14:24:09 2020

@author: hsauro
@author: joseph-hellerstein

A ModelFitter estimates parameters of a roadrunner model by using observed values
of floating species concentrations to construct fitted values with
small residuals (the difference between fitted and observed values).

Properties of interest are:
    fittedTS - estimated parameter values
    roadrunnerModel - roadrunner model object
    observedTS - observed values of floating species concentrations
    residualsTS - observedTS - fittedTS

Usage
-----
   # The constructor takes either a roadrunner or antimony model
   f = ModelFitter(model, "mydata.txt",
         parametersToFit=["k1", "k2"])
   # Fit the model parameters and view parameters
   f.fitModel()
   print(f.getFittedParameters())
   # Print observed, fitted and residual values
   print(f.observedTS)
   print(f.fittedTS)
   print(f.residualsTS)

The code is arranged as a hierarchy of classes that use the previous class:
    _modelFitterCore.ModelFitterCore - model fitting
    _modelFitterBootstrap.ModelFitterBootstrap - bootstrapping
    _modelFitterReport.ModelFitterReport - reports on results of fitting and bootstrapping
    modelFitter - plot routines
"""

from SBstoat.namedTimeseries import NamedTimeseries, TIME
import SBstoat._bootstrapResult as bsr
import SBstoat._plotOptions as po
from SBstoat._modelFitterReport import ModelFitterReport
from SBstoat.residualsAnalyzer import ResidualsAnalyzer

from docstring_expander.expander import Expander
import lmfit
import numpy as np
import pandas as pd


LOW_PERCENTILE = bsr.PERCENTILES[0]
HIGH_PERCENTILE = bsr.PERCENTILES[-1]


class ModelFitter(ModelFitterReport):

    # NOTE : The docstring for this method should contain all of the options
    #       in PlotOptions
    def plot(self, kind:str, params=None, numPoint=None, **kwargs):
        """
        High level plot routine.
        A figure may contain 1 or more plot, and each plot may contain
        1 or more curve. Figure (f), plot (p), and curve (c) are the possible
        scope of an option.  If the option is singled valued, then it applies
        to all instances of its scope. If it has multiple values, then the
        index of the value corresponds to the instance within the scope.
        A figure scope is always single valued.

        Parameters
        ----------
        kind: name of plot type
            FitAll: plots of fitted, observed data over time
            ParameterEstimatePairs: Pairwise plots of bootstrap estimates
            ParameterHistograms: histographs of bootstrap parameter values
            Residuals: plot residuals over time
            ResidualsAll: do all residuals plots
            ResidualsHistograms: histogram of residual values
            ResidualAutoCorrelations: Autocorrelations of residuals
            ResidualCrossCorrelations: Cross correlations of residuals
        alpha: (c) transparency; in [0, 1] (float)
        bins: (f) number of bins in a histogram plot (int)
        color: (c) color of the line (str, default="blue")
        columns: (f) list of columns to plot (list, default=[])
        figsize: (f) (horizontal width, vertical height) (list, default=[8, 6])
        legend: (p) tuple of str for legend (list)
        linestyle: (c) line style (str)
        marker: (c) marker for line (str)
        markersize: (c) size of marker for the line; >0 (float)
        num_row: (f) rows of plots (int)
        num_col: (f) columns of plots (int)
        subplot_width_space: (f) horizontal space between plots (float)
        timeseries2: (f) second timeseries
        title: (p) plot title (str)
        title_fontsize: (p) point size for title (float)
        title_position: (p) (x, y) relative position; x,y in [0, 1]
        suptitle: (f) figure title (str)
        xlabel: (f) x axis title (str)
        xlim: (f) order pair of lower and upper (list)
        xticklabels: (f) list of labels for x ticks (list)
        ylabel: (f) label for x axis (str)
        ylim: (f) order pair of lower and upper (str)
        yticklabels: (f) list of labels for y ticks (list)
        """
        MODULE_KINDS = [
            'ResidualsAll',
            'Residuals',
            'FitAll',
            'ParameterEstimatePairs',
            'ParameterHistograms',
            ]
        RESIDUAL_MODULE_KINDS = [
            'ResidualsHistograms',
            'ResidualAutoCorrelations',
            'ResidualCrossCorrelations',
            ]
        kinds = list(MODULE_KINDS)
        kinds.extend(RESIDUAL_MODULE_KINDS)
        if not kind in kinds:
            raise ValueError("Invalid plot kind. Options are: %s"
                             % str(kinds))
        if kind in MODULE_KINDS:
            statement = "self.plot%s(**kwargs)" % kind
            exec(statement)
        else:
            self._updateFit(params, numPoint)
            analyzer = ResidualsAnalyzer(self.observedTS, self._plotFittedTS,
                  meanFittedTS=self.bootstrapResult.fittedStatistic.meanTS,
                  stdFittedTS=self.bootstrapResult.fittedStatistic.stdTS,
                  residualsTS=self.residualsTS,
                  isPlot=self._isPlot)
            statement = "analyzer.plot%s(**kwargs)" % kind
            exec(statement)

    @Expander(po.KWARGS, po.BASE_OPTIONS, indent=8,
          header=po.HEADER)
    def plotResidualsAll(self,
          params:lmfit.Parameters=None, numPoint:int=None, **kwargs):
        """
        Plots a set of residual plots

        Parameters
        ----------
        #@expand
        """
        self._updateFit(params, numPoint)
        analyzer = ResidualsAnalyzer(self.observedTS, self._plotFittedTS,
              residualsTS=self.residualsTS,
              isPlot=self._isPlot)
        analyzer.plotAll(**kwargs)

    @Expander(po.KWARGS, po.BASE_OPTIONS, indent=8,
          header=po.HEADER)
    def plotResiduals(self,
          params:lmfit.Parameters=None, numPoint:int=None, **kwargs):
        """
        Plots residuals of a fit over time.

        Parameters
        ----------
        #@expand
        """
        self._updateFit(params, numPoint)
        analyzer = ResidualsAnalyzer(self.observedTS, self._plotFittedTS,
              residualsTS=self.residualsTS,
              isPlot=self._isPlot)
        analyzer.plotResidualsOverTime(**kwargs)

    @Expander(po.KWARGS, po.BASE_OPTIONS, indent=8, header=po.HEADER)
    def plotFitAll(self,
          params:lmfit.Parameters=None, numPoint:int=None, **kwargs):
        """
        Plots the fitted with observed data over time.

        Parameters
        ----------
        #@expand
        """
        self._updateFit(params, numPoint)
        bandLowTS = None
        bandHighTS = None
        if self.bootstrapResult is not None:
            statistic = self.bootstrapResult.simulate(numPoint=100, numSample=1000)
            self._plotFittedTS = statistic.meanTS
            bandLowTS = statistic.percentileDct[LOW_PERCENTILE]
            bandHighTS = statistic.percentileDct[HIGH_PERCENTILE]
        analyzer = ResidualsAnalyzer(self.observedTS, self._plotFittedTS,
              residualsTS=self.residualsTS,
              bandLowTS=bandLowTS, bandHighTS=bandHighTS,
              isPlot=self._isPlot)
        self._addKeyword(kwargs, po.NUM_ROW, 2)
        analyzer.plotFittedObservedOverTime(**kwargs)

    def _addKeyword(self, kwargs, key, value):
        if not key in kwargs:
            kwargs[key] = value

    def _mkParameterDF(self, parameters=None):
        df = pd.DataFrame(self.bootstrapResult.parameterDct)
        if parameters is not None:
            df = df[parameters]
        df.index.name = TIME
        return NamedTimeseries(dataframe=df)

    @Expander(po.KWARGS, po.BASE_OPTIONS, indent=8,
          header=po.HEADER)
    def plotParameterEstimatePairs(self, parameters=None, **kwargs):
        """
        Does pairwise plots of parameter estimates.

        Parameters
        ----------
        parameters: list-str
            List of parameters to do pairwise plots
        #@expand
        """
        if self.bootstrapResult is None:
            raise ValueError("Must run bootstrap before plotting parameter estimates.")
        ts = self._mkParameterDF(parameters=parameters)
        # Construct pairs
        names = list(self.bootstrapResult.parameterDct.keys())
        pairs = []
        compares = list(names)
        for name in names:
            compares.remove(name)
            pairs.extend([(name, c) for c in compares])
        #
        self._plotter.plotValuePairs(ts, pairs,
              isLowerTriangular=True, **kwargs)

    @Expander(po.KWARGS, po.BASE_OPTIONS, includes=[po.BINS], indent=8,
          header=po.HEADER)
    def plotParameterHistograms(self, parameters=None, **kwargs):
        """
        Plots histographs of parameter values from a bootstrap.

        Parameters
        ----------
        parameters: list-str
            List of parameters to do pairwise plots
        #@expand
        """
        self._checkBootstrap()
        ts = self._mkParameterDF(parameters=parameters)
        self._plotter.plotHistograms(ts, **kwargs)

    def _updateFit(self, params:lmfit.Parameters, numPoint:int):
        """
        Checks and updates the fitted values useing the specified parameters
        and number of poitns.

        Parameters
        ----------

        Returns
        -------
        """
        self._checkFit()
        if self.fittedTS is None:
            self.fittedTS = self.simulate(
                  params=params, numPoint=len(self.observedTS))
            self.residualsTS = None
        if self.residualsTS is None:
            self.residualsTS = self.fittedTS.copy()
            cols = self.observedTS.colnames
            self.residualsTS[cols] = self.observedTS[cols] - self.fittedTS[cols]
            self.residualsTS[cols]  \
                  = np.nan_to_num(self.residualsTS[cols], nan=0.0)
        self._plotFittedTS = self.simulate(params=params, numPoint=numPoint)
        self._plotFittedTS = self._plotFittedTS.subsetColumns(self.selectedColumns)



# -*- coding: utf-8 -*-
"""
Created on Tue Jul  7 14:24:09 2020

@author: hsauro
@author: joseph-hellerstein

    A TimeseriesPlotter visualizes timeseries data, especially for comparing
    time series.

    All plots are line plots with the x-axis as time. The following
    can be varied:
      single or multiple columns in a plot
      one or two timeseries plot
"""

from SBstoat import _plotOptions as po
from SBstoat import _layoutManager as lm
from SBstoat.namedTimeseries import NamedTimeseries, TIME
from SBstoat._statementManager import StatementManager

import copy
from docstring_expander.expander import Expander
import matplotlib.pyplot as plt
import numpy as np
import typing


EXPAND_KEYPHRASE = "Expansion keyphrase. Expands to help(PlotOptions()). "
EXPAND_KEYPHRASE += "Do not remove. (See timeseriesPlotter.EXPAND_KEYPRHASE.)"
LABEL1 = "1"
LABEL2 = "2"
COLORS = ['r', 'g', 'b', 'c', 'pink', 'grey']
NULL_STR = ""
# Ensure lots of colors
COLORS.extend(COLORS)
COLORS.extend(COLORS)
# Title positions
POS_MID = 0.5
POS_TOP = 0.9
# Autocorrelations
MAX_LAGS = 10
NUM_STD = 2
#
IDX_PLOT1 = 0  # plot index for the first plot
IDX_PLOT2 = 1  # plot index for the second plot
DEFAULT_MARKER = "o"


########################################
class TimeseriesPlotter():

    def __init__(self, isPlot=True):
        """
        Parameters
        ---------
        isPlot: boolean
            display the plot

        Example
        -------
        plotter = TimeseriesPlotter(timeseries)
        plotter.plotTimeSingle()  # Single variable plots for one timeseries
        """
        self.isPlot = isPlot

    @staticmethod
    def _mkPlotOptionsMatrix(timeseries1, maxCol=None, **kwargs):
        """
        Creates PlotOptions for a dense matrix of plots.

        Parameters
        ----------
        timeseries1: NamedTimeseries
        isLowerTriangular: bool
            is a lower triangular matix
        maxCol: int
            maximum number of columns
        kwargs: dict

        Returns
        -------
        PlotOptions
            assigns values to options.numRow, options.numCol
        """
        options = po.PlotOptions()
        # bins
        if po.BINS in kwargs.keys():
            bins = kwargs[po.BINS]
        else:
            bins = None
        options.bins = bins
        # Second timeseries
        if po.TIMESERIES2 in kwargs.keys():
            options.timeseries2 = kwargs[po.TIMESERIES2]
        if maxCol is None:
            if po.COLUMNS in kwargs:
                maxCol = len(kwargs[po.COLUMNS])
            else:
                maxCol = len(timeseries1.colnames)
        # States
        hasRow, hasCol = TimeseriesPlotter._getOptionValueState(kwargs, options)
        # Assignments based on state
        if hasRow and hasCol:
            # Have been assigned
            pass
        elif hasRow and (not hasCol):
            options.numCol = int(maxCol/options.numRow)
        elif (not hasRow) and hasCol:
            options.numRow = int(maxCol/options.numCol)
        else:
            options.numRow = 1
            options.numCol = maxCol
        if maxCol > options.numRow*options.numCol:
            options.numRow += 1
        options.initialize(timeseries1, **kwargs)
        #
        return options

    @staticmethod
    def _getOptionValueState(kwargs, options):
        hasCol = False
        hasRow = False
        if po.NUM_ROW in kwargs:
            options.numRow = kwargs[po.NUM_ROW]
            hasRow = True
        if po.NUM_COL in kwargs:
            options.numCol = kwargs[po.NUM_COL]
            hasCol = True
        return hasRow, hasCol

    @staticmethod
    def _mkPlotOptionsLowerTriangular(timeseries1, numPlot,  **kwargs):
        """
        Creates PlotOptions for a lower triangular matrix of plots.
        2*len(pairs) - N = N**2

        Parameters
        ----------
        timeseries1: NamedTimeseries
        numPlot: int
        kwargs: dict

        Returns
        -------
        PlotOptions
            assigns values to options.numRow, options.numCol
        """
        options = po.PlotOptions()
        if numPlot == 1:
            options.numRow = 1
            options.numCol = 1
        options.initialize(timeseries1, **kwargs)
        #
        return options

    @Expander(po.KWARGS, po.BASE_OPTIONS, excludes=[po.TITLE], indent=8,
           header=po.HEADER)
    def plotTimeSingle(self, timeseries1:NamedTimeseries,
           meanTS:NamedTimeseries=None, stdTS:NamedTimeseries=None,
           bandLowTS:NamedTimeseries=None, bandHighTS:NamedTimeseries=None,
           ax_spec=None,
           position=None,  **kwargs):
        """
        Constructs plots of single columns, possibly with a second
        timeseries.

        Parameters
        ---------
        timeseries1: timeseries to plot
        meanTS: mean values of timeseries to plot
        stdTS: standard deviations of meanTS (same times)
        bandLowTS: timeseries that describes the lower band for timeseries1
        bandhighTS: timeseries that describes the higher band for timeseries1
        ax_spec: Optionally specified axis for all lines
        position: (int, int) - position of ax_spec
        #@expand

        Example
        -------
        plotter = TimeseriesPlotter()
        plotter.plotTimeSingle(timeseries)
        """
        options = TimeseriesPlotter._mkPlotOptionsMatrix(timeseries1, **kwargs)
        # Adjust rows and columns
        numPlot = len(options.columns)  # Number of plots
        if po.NUM_COL in kwargs.keys():
            options.numRow = int(np.ceil(numPlot/options.numCol))
        else:
            options.numCol = int(np.ceil(numPlot/options.numRow))
        options.set(po.XLABEL, TIME)
        # Create the LayoutManager
        if ax_spec is None:
            layout = self._mkManager(options, numPlot)
        else:
            layout = None
        #
        def mkStatement(ax, statement:StatementManager,
            timeseries:NamedTimeseries, variable:str):
            """
            Creates a plotting statement.

            Parameters
            ----------
            ax: plotting axes
            initialStatement: statement if initialized
            timeseries: what to plot
            variable: variable to plot

            Returns
            -------
            StatementManager
            """
            times, values = self._adjustSeriesForNan(
                  timeseries[TIME], timeseries[variable])
            statement.addPosarg(times)
            statement.addPosarg(values)
            return statement
        #
        def isFirstColumn(plotIdx):
            if layout is not None:
                return layout.isFirstColumn(plotIdx)
            if position is not None:
                return position[1]  == 0
            return True
        #
        def isLastRow(plotIdx):
            if layout is not None:
                return layout.isLastRow(plotIdx)
            if po.NUM_ROW in kwargs.keys():
                numRow = kwargs[po.NUM_ROW]
            else:
                return True
            if position is None:
                return True
            return position[0] + 1 == numRow
        #
        def mkBand(ts, variable, plotIdx, lineIdx):
            initialStatement = StatementManager(ax.fill_between)
            statement = mkStatement(ax, initialStatement, timeseries1, variable)
            statement.addPosarg(ts[variable])
            statement.addKwargs(alpha=0.2)
            options.do(ax, statement=statement, plotIdx=plotIdx, lineIdx=lineIdx)
        #
        # Construct the plots
        baseOptions = copy.deepcopy(options)
        for plotIdx, variable in enumerate(options.columns):
            if ax_spec is None:
                ax = layout.getAxis(plotIdx)
            else:
                ax = ax_spec
            options = copy.deepcopy(baseOptions)
            #ax = axes[row, col]
            options.set(po.YLABEL, "concentration")
            options.set(po.TITLE, variable)
            if not isFirstColumn(plotIdx):
                options.ylabel =  NULL_STR
                options.set(po.YLABEL, "", isOverride=True)
            if not isLastRow(plotIdx):
                options.set(po.XLABEL, "", isOverride=True)
            # Construct the plot
            lineIdx = 0
            initialStatement = StatementManager(ax.plot)
            statement = mkStatement(ax, initialStatement, timeseries1, variable)
            options.do(ax, statement=statement, plotIdx=plotIdx, lineIdx=lineIdx)
            legends = []
            if options.timeseries2 is not None:
                lineIdx = 1
                initialStatement = StatementManager(ax.plot)
                if options.marker is not None:
                    if len(options.marker) > 1:
                        if options.marker[lineIdx] is not None:
                            initialStatement = StatementManager(ax.scatter)
                    elif options.marker is not None:
                        initialStatement = StatementManager(ax.scatter)
                statement = mkStatement(ax, initialStatement,
                      options.timeseries2, variable)
                legends = [LABEL1, LABEL2]
                options.lengends = legends
                options.do(ax, statement=statement, plotIdx=plotIdx, lineIdx=lineIdx)
            if  meanTS is not None:
                if stdTS is None:
                    stdTS = meanTS.copy(isInitialize=True)
                lineIdx = 2
                initialStatement = StatementManager(ax.scatter)
                statement = mkStatement(ax, initialStatement, meanTS, variable)
                options.marker = "^"
                statement.addKwargs(facecolors="none")
                options.do(ax, statement=statement, plotIdx=plotIdx, lineIdx=lineIdx)
                #
                initialStatement = StatementManager(ax.errorbar)
                statement = mkStatement(ax, initialStatement, meanTS, variable)
                statement.addKwargs(yerr=stdTS[variable])
                statement.addKwargs(linestyle="none")
                legends.append("Bootstrap mean")
                options.lengends = legends
                options.do(ax, statement=statement, plotIdx=plotIdx, lineIdx=lineIdx)
            if  bandLowTS is not None:
                lineIdx = 0
                mkBand(bandLowTS, variable, plotIdx, lineIdx)
            if  bandHighTS is not None:
                lineIdx = 0
                mkBand(bandHighTS, variable, plotIdx, lineIdx)
        if self.isPlot:
            plt.show()

    @Expander(po.KWARGS, po.BASE_OPTIONS, excludes=[po.NUM_COL, po.NUM_ROW],
          indent=8, header=po.HEADER)
    def plotTimeMultiple(self, timeseries1, **kwargs):
        """
        Constructs a plot with all columns in a timeseries.
        If there is a second timeseries, then there are 2 plots.

        Parameters
        ---------
        timeseries1: NamedTimeseries
        #@expand
        """
        numPlot = 2 if po.TIMESERIES2 in kwargs else 1
        maxCol = numPlot
        options = TimeseriesPlotter._mkPlotOptionsMatrix(timeseries1, maxCol=maxCol, **kwargs)
        if po.COLOR not in kwargs.keys():
            colors = list(COLORS)
            while len(colors) < len(timeseries1.colnames):
                colors.extend(COLORS)
            options.color = colors
        #
        def multiPlot(ax:plt.Axes, timeseries:NamedTimeseries,
              options:po.PlotOptions, plotIdx:int):
            if plotIdx == IDX_PLOT2:
                timeseries = options.timeseries2
            for lineIdx, col in enumerate(timeseries.colnames):
                if isinstance(options.marker, list):
                    marker = options.marker[lineIdx]
                else:
                    marker = options.marker
                if marker is None:
                    statement = StatementManager(ax.plot)
                else:
                    statement = StatementManager(ax.scatter)
                times, values = self._adjustSeriesForNan(
                      timeseries[TIME], timeseries[col])
                statement.addPosarg(times)
                statement.addPosarg(values)
                options.legend = timeseries.colnames
                options.do(ax, statement=statement, plotIdx=plotIdx, lineIdx=lineIdx)
        #
        # Update rows and columns
        if options.timeseries2 is None:
            options.numRow = 1
            options.numCol = 1
        else:
            if po.NUM_ROW in kwargs:
                options.numCol = int(np.ceil(2/options.numRow))
            else:
                options.numRow = int(np.ceil(2/options.numCol))
        # Create the LayoutManager
        layout = self._mkManager(options, numPlot)
        # Construct the plots
        options.set(po.XLABEL, TIME)
        ax = layout.getAxis(0)
        multiPlot(ax, timeseries1, options, IDX_PLOT1)
        if options.timeseries2 is not None:
            ax = layout.getAxis(numPlot-1)
            multiPlot(ax, options.timeseries2, options, IDX_PLOT2)
        if self.isPlot:
            plt.show()

    def _mkManager(self, options, numPlot, isLowerTriangular=False):
        if numPlot == 1:
            layout = lm.LayoutManagerSingle(
                   options, numPlot)
        elif isLowerTriangular:
            options.set(po.TITLE_POSITION,
                (POS_MID, POS_TOP))
            layout = lm.LayoutManagerLowerTriangular(
                   options, numPlot)
        else:
            options.set(po.TITLE_POSITION,
                (POS_MID, POS_TOP))
            layout = lm.LayoutManagerMatrix(options,
                  numPlot)
        return layout

    @Expander(po.KWARGS, po.BASE_OPTIONS,
          excludes=[po.NUM_ROW, po.NUM_COL], indent=8,
          header=po.HEADER)
    def plotValuePairs(self, timeseries, pairs,
              isLowerTriangular=False, **kwargs):
        """
        Constructs plots of values of column pairs for a single
        timeseries.

        Parameters
        ---------
        timeseries: NamedTimeseries
        #@expand
        """
        numPlot = len(pairs)
        if isLowerTriangular:
            options = TimeseriesPlotter._mkPlotOptionsLowerTriangular(
                  timeseries, numPlot, **kwargs)
        else:
            options = TimeseriesPlotter._mkPlotOptionsMatrix(timeseries,
                  maxCol=numPlot, **kwargs)
        if options.marker is None:
            options.marker = DEFAULT_MARKER
        layout = self._mkManager(options, numPlot,
              isLowerTriangular=isLowerTriangular)
        options.xlabel = NULL_STR
        baseOptions = copy.deepcopy(options)
        for plotIdx, pair in enumerate(pairs):
            options = copy.deepcopy(baseOptions)
            xVar = pair[0]
            yVar = pair[1]
            ax = layout.getAxis(plotIdx)
            statement = StatementManager(ax.scatter)
            statement.addPosarg(timeseries[xVar])
            statement.addPosarg(timeseries[yVar])
            if isLowerTriangular:
                if layout.isLastRow(plotIdx):
                    options.xlabel = xVar
                else:
                    options.title = NULL_STR
                    options.xticklabels = []
                if layout.isFirstColumn(plotIdx):
                    options.ylabel = yVar
                else:
                    options.ylabel = NULL_STR
                    options.yticklabels = []
            else:
                # Matrix plot
                options.xlabel = NULL_STR
                options.ylabel = NULL_STR
                options.title = "%s v. %s" % (xVar, yVar)
            options.do(ax, statement=statement, plotIdx=plotIdx)
        if self.isPlot:
            plt.show()

    @Expander(po.KWARGS, po.BASE_OPTIONS, includes=[po.BINS], indent=8,
          header=po.HEADER)
    def plotHistograms(self, timeseries, **kwargs):
        """
        Constructs a matrix of histographs for timeseries values.

        Parameters
        ---------
        timeseries: NamedTimeseries
        #@expand
        """
        options = TimeseriesPlotter._mkPlotOptionsMatrix(timeseries, **kwargs)
        numPlot = len(options.columns)
        layout = self._mkManager(options, numPlot)
        baseOptions = copy.deepcopy(options)
        for plotIdx, column in enumerate(options.columns):
            options = copy.deepcopy(baseOptions)
            ax = layout.getAxis(plotIdx)
            if layout.isFirstColumn(plotIdx):
                options.ylabel = "density"
            if not layout.isFirstColumn(plotIdx):
                options.ylabel = NULL_STR
            if layout.isLastRow(plotIdx):
                options.xlabel = "value"
            else:
                options.xlabel = NULL_STR
            # Matrix plot
            options.title = column
            statement = StatementManager(ax.hist)
            statement.addPosarg(timeseries[column])
            statement.addKwargs(density=True)
            if options.bins is not None:
                statement.addKwargs(bins=options.bins)
            options.do(ax, statement=statement, plotIdx=plotIdx)
        if self.isPlot:
            plt.show()

    @Expander(po.KWARGS, po.BASE_OPTIONS, indent=8,
          header=po.HEADER)
    def plotCompare(self,
           ts1: NamedTimeseries,
           ts2: NamedTimeseries,
           **kwargs: dict):
        """
        Plots columns against each other.

        Parameters
        ----------
        #@expand
        """
        mergedTS = ts1.concatenateColumns(ts2)
        pairs = [(c, "%s_" % c) for c in ts1.colnames]
        #
        self.plotValuePairs(mergedTS, pairs, **kwargs)

    def _mkAutocorrelation(self, timeseries:NamedTimeseries, numPlot:int=None,
          isLowerTriangular:bool=False, **kwargs:dict)  \
          -> typing.Tuple[po.PlotOptions, lm.LayoutManager]:
        """
        Constructs options and layout for autocorrelation plots.
        """
        if isLowerTriangular:
            options = TimeseriesPlotter._mkPlotOptionsLowerTriangular(
                  timeseries, numPlot, **kwargs)
        else:
            options = TimeseriesPlotter._mkPlotOptionsMatrix(timeseries, **kwargs)
            numPlot = len(options.columns)
        layout = self._mkManager(options, numPlot,
              isLowerTriangular=isLowerTriangular)
        options.set(po.XLABEL, "lag")
        options.set(po.YLABEL, "autocorrelation")
        options.set(po.YLIM, [-1.2, 1.2])
        return options, layout

    def _mkAutocorrelationErrorBounds(self, timeseries:NamedTimeseries)  \
          -> typing.Tuple[typing.List[int], typing.List[float], typing.List[float]]:
        def mkFullArray(arr):
            reverse_arr = arr[::-1]
            return np.concatenate([reverse_arr, arr[1:]])
        # Construct lags
        upper_lags = np.array(range(1, MAX_LAGS+1))
        lower_lags = -1*upper_lags
        lower_lags.sort()
        lags = np.concatenate([lower_lags, np.array([0]), upper_lags])
        #
        length = len(timeseries)
        upper_line = np.array([NUM_STD/(np.sqrt(length - l))
              for l in range(MAX_LAGS+1)])
        lower_line = -1*upper_line
        upper_line = mkFullArray(upper_line)
        lower_line = mkFullArray(lower_line)
        #
        return lags, lower_line, upper_line

    @Expander(po.KWARGS, po.BASE_OPTIONS, indent=8,
          header=po.HEADER)
    def plotAutoCorrelations(self, timeseries:NamedTimeseries,
           **kwargs:dict):
        """
        Plots autocorrelations for a timeseries.
        Uses the Barlet bound to estimate confidence intervals.

        Parameters
        ----------
        #@expand
        """
        # Structure the plots
        baseOptions, layout = self._mkAutocorrelation(timeseries,
              isLowerTriangular=False, **kwargs)
        lags, lower_line, upper_line = self._mkAutocorrelationErrorBounds(timeseries)
        baseOptions.set(po.SUPTITLE, "Autocorrelations")
        for plotIdx, column in enumerate(baseOptions.columns):
            options = copy.deepcopy(baseOptions)
            ax = layout.getAxis(plotIdx)
            if not layout.isFirstColumn(plotIdx):
                options.ylabel = NULL_STR
            if not layout.isLastRow(plotIdx):
                options.xlabel = NULL_STR
            # Matrix plot
            lineIdx = 0
            options.title = column
            statement = StatementManager(ax.acorr)
            statement.addPosarg(timeseries[column])
            statement.addKwargs(maxlags=MAX_LAGS)
            options.do(ax, statement=statement, plotIdx=plotIdx, lineIdx=lineIdx)
            #
            self._addConfidenceLines(ax, lags, [lower_line, upper_line], options,
                  lineIdx, plotIdx=plotIdx)
        if self.isPlot:
            plt.show()

    def _addConfidenceLines(self, ax, xline, lines, options, startLineIdx, **kwargs):
        lineIdx = startLineIdx
        for line in lines:
            lineIdx += 1
            statement = StatementManager(ax.plot)
            statement.addPosarg(xline)
            statement.addPosarg(line)
            options.do(ax, statement=statement, lineIdx=lineIdx, **kwargs)

    @Expander(po.KWARGS, po.BASE_OPTIONS, indent=8,
          header=po.HEADER)
    def plotCrossCorrelations(self, timeseries:NamedTimeseries,
          **kwargs:dict):
        """
        Constructs pairs of cross correlation plots.
        Uses the Barlet bound to estimate confidence intervals.

        Parameters
        ----------
        #@expand
        """
        if po.COLUMNS in kwargs.keys():
            numCol = len(po.COLUMNS)
        else:
            numCol = len(timeseries.colnames)
        numPlot = int((numCol**2 - numCol)/2)
        baseOptions, layout = self._mkAutocorrelation(timeseries,
              numPlot=numPlot, isLowerTriangular=True, **kwargs)
        # Construct the plot pairs
        pairs = []
        columns = list(baseOptions.columns)
        for col1 in baseOptions.columns:
            columns.remove(col1)
            for col2 in columns:
                pairs.append((col1, col2))
        lags, lower_line, upper_line = self._mkAutocorrelationErrorBounds(
              timeseries)
        baseOptions.set(po.SUPTITLE, "Cross Correlations")
        # Do the plots
        for plotIdx, pair in enumerate(pairs):
            options = copy.deepcopy(baseOptions)
            xVar = pair[0]
            yVar = pair[1]
            options.set(po.TITLE, "%s, %s" % (xVar, yVar))
            ax = layout.getAxis(plotIdx)
            if layout.isLastRow(plotIdx):
                options.xlabel = "lag"
            else:
                options.xticklabels = []
            if layout.isFirstColumn(plotIdx):
                options.ylabel = "corr"
            else:
                options.ylabel = NULL_STR
                options.yticklabels = []
            # Matrix plot
            lineIdx = 0
            statement = StatementManager(ax.xcorr)
            statement.addPosarg(timeseries[xVar])
            statement.addPosarg(timeseries[yVar])
            statement.addKwargs(maxlags=MAX_LAGS)
            options.do(ax, statement=statement, plotIdx=plotIdx, lineIdx=lineIdx)
            self._addConfidenceLines(ax, lags, [lower_line, upper_line], options,
                  lineIdx, plotIdx=plotIdx)
        if self.isPlot:
            plt.show()

    def _adjustSeriesForNan(self, times, values):
        indices = [i for i in range(len(values)) if not np.isnan(values[i])]
        return times[indices], values[indices]



# -*- coding: utf-8 -*-
"""
Created on Aug 2020

@author: joseph-hellerstein

Manages the layout of a matrix of plots.
    LayoutManager - abstract class
    LayoutManagerSingle - single plot
    LayoutManagerMatrix - dense matrix of plots
    LayoutManagerLowerTriangular - plots in lower triangle
"""

import matplotlib.pyplot as plt
import numpy as np


########################################
class LayoutManager():
    """
    Manages the position of plots on the figure by creating
    a mapping from a linear sequence of plots to positions
    in a 2-dimensional grid.

    The top level class is abstract. Subclasses must override
    the method _setAxes that mplements the above mapping.
    """

    def __init__(self, options, numPlot):
        """
        Manages allocation of space in subplot and sets up options
        accordingly.

        Parameters
        ---------
        options: PlotOptions
        """
        self.options = options
        self.numPlot = numPlot
        # Number of plot positions in the figure
        if (self.options.numRow is not None)   \
              and (self.options.numCol is not None):
            self.numPlotPosition = self.options.numRow*self.options.numCol
        else:
            self.numPlotPosition = None
        self.figure, self.axes, self.axisPositions = self._setAxes()

    def _initializeAxes(self):
        """
        Common codes for setAxes() methods.

        Returns
        -------
        Maplotlib.Figure, Matplotlib.Axes
        """
        figure = plt.figure(figsize=self.options.figsize)
        axes = []
        plt.subplots_adjust(wspace=0.5)
        return figure, axes

    def _setAxes(self):
        """
        Implements the mapping from a linear sequence to positions in
        a two dimension grid for the figure.
        Must override.

        Returns
        -------
        Maplotlib.Figure, Matplotlib.Axes, list-tuple
        """
        raise RuntimeError("Must override.")

    def _calcRowColumn(self, index):
        """
        Returns row and column with 0 indexing for plot coordinates in figure.

        Parameters
        ----------
        index: int

        Returns
        -------
        int, int
        """
        row = index // self.options.numCol
        col = index % self.options.numCol
        return row, col

    def getAxis(self, index):
        """
        Returns the axis to use for the plot index.

        Parameters
        ----------
        index: int

        Returns
        -------
        Matplotlib.axes
        """
        return self.axes[index]

    def isFirstColumn(self, index):
        _, col = self.axisPositions[index]
        return col == 0

    def isFirstRow(self, index):
        row, _ = self.axisPositions[index]
        return row == 0

    def isLastCol(self, index):
        _, col = self.axisPositions[index]
        return col == self.options.numCol - 1

    def isLastRow(self, index):
        row, _ = self.axisPositions[index]
        return row == self.options.numRow - 1


########################################
class LayoutManagerSingle(LayoutManager):
    """
    LayoutManager for a figure consisting of one plot.
    """

    def _setAxes(self):
        """
        len(self.axes) == 1

        Returns
        -------
        Maplotlib.Figure, Matplotlib.Axes
        """
        figure, _ = self._initializeAxes()
        axes = np.array([plt.subplot(1, 1, 1)])
        return figure, axes, [(0, 0)]


########################################
class LayoutManagerMatrix(LayoutManager):
    """
    LayoutManager for a figure that is a matrix
    of plots, although not all entries need be present.
    """

    def _setAxes(self):
        """
        Linearizes the matrix counting across rows and then down
        by column.

        Returns
        -------
        Maplotlib.Figure, Matplotlib.Axes
        """
        figure, axes = self._initializeAxes()
        axisPositions = []
        for index in range(self.numPlotPosition):
            row, col = self._calcRowColumn(index)
            if index < self.numPlot:
                ax = plt.subplot2grid( (self.options.numRow,
                      self.options.numCol), (row, col))
                axisPositions.append((row, col))
            axes.append(ax)
        return figure, axes, axisPositions


####
class LayoutManagerLowerTriangular(LayoutManager):
    """
    LayoutManager for a figure that is an upper triangular plot.
    """

    def __init__(self, options, numPlot):
        """
        Manages allocation of space in subplot and sets up options
        accordingly.

        Parameters
        ---------
        options: PlotOptions
        """
        self.options = options
        # Adjust rows and columns for the number of plots
        # by creating a square matrix with the right number of
        # plot positions to accommodate the number of plots.
        matSize = None
        for mm in range(numPlot**3):
            if 2*numPlot - mm <= mm**2:
                matSize = mm
                break
        if matSize is None:
            raise RuntimeError("Could not find matrix size.")
        # Number of plot positions in the figure
        self.options.numRow = matSize
        self.options.numCol = matSize
        self.numPlotPosition = self.options.numRow*self.options.numCol
        self.figure, self.axes, self.axisPositions = self._setAxes()

    def _setAxes(self):
        """
        Linearizes the lower triangular matrix counting down successive columns
        starting at row=col.

        Returns
        -------
        Maplotlib.Figure, Matplotlib.Axes
        """
        figure, axes = self._initializeAxes()
        axisPositions = []
        row = 0
        col = 0
        for _ in range(self.numPlotPosition):
            if row >= col:
                ax = plt.subplot2grid(
                      (self.options.numRow, self.options.numCol), (row, col))
                axes.append(ax)
                axisPositions.append((row, col))
            # Update row, col
            if row == self.options.numRow - 1:
                row = 0
                col += 1
            else:
                row += 1
        return figure, axes, axisPositions



# -*- coding: utf-8 -*-
"""
 Created on August 18, 2020

@author: joseph-hellerstein

Reports for model fitter
"""

from SBstoat.namedTimeseries import NamedTimeseries, TIME, mkNamedTimeseries
from SBstoat._modelFitterBootstrap import ModelFitterBootstrap
from SBstoat import _helpers

import lmfit
import numpy as np
import pandas as pd
import typing


##############################
class ModelFitterReport(ModelFitterBootstrap):

    def reportFit(self)->str:
        """
        Provides details of the parameter fit.
        Notes:
            1. Deletes lines about the optimization performance since the


        Example
        -------
        f.reportFit()
        """
        VARIABLE_STG = "[[Variables]]"
        CORRELATION_STG = "[[Correlations]]"
        self._checkFit()
        if self.minimizerResult is None:
            raise ValueError("Must do fitModel before reportFit.")
        valuesDct = self.params.valuesdict()
        valuesStg = _helpers.ppDict(dict(valuesDct), indent=4)
        reportSplit = str(lmfit.fit_report(self.minimizerResult)).split("\n")
        # Eliminate Variables section
        inVariableSection = False
        trimmedReportSplit = []
        for line in reportSplit:
            if VARIABLE_STG in line:
                inVariableSection = True
            if CORRELATION_STG in line:
                inVariableSection = False
            if inVariableSection:
                continue
            else:
                trimmedReportSplit.append(line)
        # Construct the report
        newReportSplit = [VARIABLE_STG]
        newReportSplit.extend(valuesStg.split("\n"))
        newReportSplit.extend(trimmedReportSplit)
        return "\n".join(newReportSplit)

    def reportBootstrap(self):
        """
        Prints a report of the bootstrap results.
        ----------

        Example
        -------
        f.reportBootstrap()
        """
        self._checkBootstrap()
        self.logger.activity(self.bootstrapResult)
